# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import operator
import sys

import numpy as np
import pytest
import scipy as sc

from sisl._array import array_arange
from sisl._core.sparse import *
from sisl._core.sparse import indices

pytestmark = [
    pytest.mark.sparse,
    pytest.mark.filterwarnings("ignore", category=sc.sparse.SparseEfficiencyWarning),
]


@pytest.fixture
def s1():
    return SparseCSR((10, 100), dtype=np.int32)


@pytest.fixture
def s1d():
    return SparseCSR((10, 100))


@pytest.fixture
def s2():
    return SparseCSR((10, 100, 2))


def test_indices():
    search = np.array([1, 4, 5, 0], np.int32)
    val = np.array([0, 5], np.int32)
    idx = indices(search, val, 0)
    assert np.allclose(idx, [3, 2])
    val = np.array([0, 1, 5], np.int32)
    idx = indices(search, val, 0)
    assert np.allclose(idx, [3, 0, 2])
    idx = indices(search, val, 20)
    assert np.allclose(idx, [23, 20, 22])
    val = np.array([-1, 0, 1, 5, 10], np.int32)
    idx = indices(search, val, 20)
    assert np.allclose(idx, [-1, 23, 20, 22, -1])


def test_fail_init1():
    with pytest.raises(ValueError):
        SparseCSR((10, 100, 20, 20), dtype=np.int32)


def test_fail_init_shape0():
    with pytest.raises(ValueError):
        SparseCSR((0, 10, 10), dtype=np.int32)


def test_fail_init2():
    data = np.empty([2, 2], np.float64)
    indices = np.arange(2)
    indptr = np.arange(2)
    with pytest.raises(ValueError):
        SparseCSR((data, indices, indptr, indptr), shape=(100, 20, 20))


def test_init_csr_inputs():
    data = np.empty([2, 2], np.float64)
    indices = np.arange(2)
    indptr = np.arange(3)
    SparseCSR((data, indices, indptr))


def test_csr_matrix_int64_fail():
    # fixes #901
    s = sc.sparse.csr_matrix((2, 2))
    s[0, 0] = 1
    s[1, 1] = 1
    # this will also break scipy, but here for testing purposes:
    s.indices = s.indices.astype(np.int64)
    s.indptr = s.indptr.astype(np.int64)
    sp = SparseCSR.fromsp(s)
    assert sp.col.dtype == np.int32
    assert sp.ncol.dtype == np.int32
    assert sp.ptr.dtype == np.int32


def test_fail_align1():
    s1 = SparseCSR((10, 100), dtype=np.int32)
    s2 = SparseCSR((20, 100), dtype=np.int32)
    with pytest.raises(ValueError):
        s1.align(s2)


def test_set_get1():
    s1 = SparseCSR((10, 10, 3), dtype=np.int32)
    s1[1, 1] = [1, 2, 3]
    s1[1, 2] = [1, 2, 3]
    assert s1.nnz == 2
    assert np.all(s1[1, 1] == [1, 2, 3])
    assert np.all(s1[1, [1, 2]] == [[1, 2, 3], [1, 2, 3]])
    assert np.all(s1[1, [1, 2], 1] == [2, 2])
    assert np.all(s1[1, [1, 6], 1] == [2, 0])


def test_init1(s1, s2, s1d):
    str(s1)
    assert s1.dtype == np.int32
    assert s1d.dtype == np.float64
    assert s2.dtype == np.float64


def test_init2():
    SparseCSR((10, 100))
    for d in [np.int32, np.float64, np.complex128]:
        s = SparseCSR((10, 100), dtype=d)
        assert s.shape, (10, 100 == 1)
        assert s.dim == 1
        assert s.dtype == d
        for k in [1, 2]:
            s = SparseCSR((10, 100, k), dtype=d)
            assert s.shape, (10, 100 == k)
            assert s.dim == k
            s = SparseCSR((10, 100), dim=k, dtype=d)
            assert s.shape, (10, 100 == k)
            assert s.dim == k
            s = SparseCSR((10, 100, 3), dim=k, dtype=d)
            assert s.shape, (10, 100 == 3)
            assert s.dim == 3


def test_init3():
    lil = sc.sparse.lil_matrix((10, 10), dtype=np.int32)
    lil[0, 1] = 1
    lil[0, 2] = 2
    sp = SparseCSR(lil)
    assert sp.dtype == np.int32
    assert sp.shape, (10, 10 == 1)
    assert sp.nnz == 2
    assert sp[0, 1] == 1
    assert sp[0, 2] == 2
    sp = SparseCSR(lil, dtype=np.float64)
    assert sp.shape, (10, 10 == 1)
    assert sp.dtype == np.float64
    assert sp.nnz == 2
    assert sp[0, 1] == 1
    assert sp[0, 2] == 2


def test_init4():
    lil = sc.sparse.lil_matrix((10, 10), dtype=np.int32)
    lil[0, 1] = 1
    lil[0, 2] = 2
    csr = lil.tocsr()
    sp = SparseCSR((csr.data, csr.indices, csr.indptr))
    assert sp.dtype == np.int32
    assert sp.shape, (10, 10 == 1)
    assert sp.nnz == 2
    assert sp[0, 1] == 1
    assert sp[0, 2] == 2
    sp = SparseCSR((csr.data, csr.indices, csr.indptr), dtype=np.float64)
    assert sp.shape, (10, 10 == 1)
    assert sp.dtype == np.float64
    assert sp.nnz == 2
    assert sp[0, 1] == 1
    assert sp[0, 2] == 2


def test_diagonal_1d():
    s = SparseCSR((10, 100), dtype=np.int32)
    for i in range(s.shape[0]):
        s[i, i] = i
    d = s.diagonal()
    assert np.allclose(d, np.arange(s.shape[0]))


def test_diagonal_2d():
    s = SparseCSR((10, 100, 2), dtype=np.int32)
    d1 = np.arange(s.shape[0])
    d2 = np.arange(s.shape[0]) + 4

    for i in range(s.shape[0]):
        s[i, i] = (d1[i], d2[i])
    d = s.diagonal()

    assert np.allclose(d[:, 0], d1)
    assert np.allclose(d[:, 1], d2)


def test_extend1():
    csr = SparseCSR((10, 10), nnzpr=1, dtype=np.int32)
    csr[1, 1] = 3
    csr[2, 2] = 4
    csr[0, 1] = 1
    csr[0, 2] = 2
    csr[0, []] = 2
    assert csr[0, 1] == 1
    assert csr[0, 2] == 2
    assert csr[1, 1] == 3
    assert csr[2, 2] == 4
    assert csr.nnz == 4
    assert np.allclose(csr[0, [0, 1, 2]], [0, 1, 2])


def test_diags_1():
    csr = SparseCSR((10, 10), nnzpr=1, dtype=np.int32)
    csr1 = csr.diags(10)
    assert csr1.shape[0] == 10
    assert csr1.shape[1] == 10
    assert csr1.shape[2] == 1
    assert csr1.nnz == 10
    assert np.allclose(csr1.toarray()[..., 0], np.diag(np.full(10, 10)))
    csr2 = csr.diags(10)
    csr3 = csr.diags(np.zeros(10, np.int32))
    assert csr2.spsame(csr3)


def test_diags_2():
    csr = SparseCSR((10, 10), dim=3, nnzpr=1, dtype=np.int32)
    csr1 = csr.diags(10, dim=1)
    csr2 = csr.diags(10, dim=2, dtype=np.int16)
    assert csr1.shape[2] == 1
    assert csr2.shape[2] == 2
    assert csr1.nnz == 10
    assert csr2.nnz == 10
    assert csr1.spsame(csr2)
    # TODO fix diags such that it won't affect
    assert csr1.dtype == np.int64
    assert csr2.dtype == np.int16


def test_diags_offsets():
    csr = SparseCSR((10, 10), dtype=np.int32)
    m = csr.diags(1)
    assert m.nnz == 10
    assert np.sum(m) == 10
    m = csr.diags(1, offsets=1)
    assert m.nnz == 9
    assert np.sum(m) == 9
    m = csr.diags(1, offsets=2)
    assert m.nnz == 8
    assert np.sum(m) == 8
    m = csr.diags(1, offsets=-2)
    assert m.nnz == 8
    assert np.sum(m) == 8


def test_diags_multiple_diagonals():
    csr = SparseCSR((10, 10), dim=2, dtype=np.int32)
    m = csr.diags([[1, 2]])
    assert np.allclose(m[0, 0], [1, 2])


def test_create1(s1d):
    s1d[0, [1, 2, 3]] = 1
    assert s1d.nnz == 3
    s1d[2, [1, 2, 3]] = 1
    assert s1d.nnz == 6
    s1d.empty(keep_nnz=True)
    assert s1d.nnz == 6
    s1d.empty()
    assert s1d.nnz == 0
    s1d[0, 0] = np.nan
    # nan produces zeros
    assert s1d.nnz == 1
    s1d.empty()


def test_create2(s1):
    assert len(s1) == s1.shape[0]
    for i in range(10):
        j = range(i * 4, i * 4 + 3)
        s1[0, j] = i
        assert s1.nnz == (i + 1) * 3
        for jj in j:
            assert s1[0, jj] == i
            assert s1[1, jj] == 0


def test_create3(s1):
    for i in range(10):
        j = range(i * 4, i * 4 + 3)
        s1[0, j] = i
        assert s1.nnz == (i + 1) * 3
        s1[0, range((i + 1) * 4, (i + 1) * 4 + 3)] = None
        assert s1.nnz == (i + 1) * 3
        for jj in j:
            assert s1[0, jj] == i
            assert s1[1, jj] == 0


def test_create_1d_bcasting_data_1d(s1):
    s2 = s1.copy()
    for i in range(10):
        j = range(i * 4, i * 4 + 3)
        s2[0, j] = i
        s2[1, j] = i
        s2[2, j] = i
        s2[3, j] = i

    s3 = s1.copy()
    for i in range(10):
        j = np.arange(i * 4, i * 4 + 3).reshape(1, -1)
        s3[np.arange(4).reshape(-1, 1), j] = i

    assert s2.spsame(s3)
    assert np.sum(s2 - s3) == 0


@pytest.mark.xfail(
    sys.platform.startswith("win"), reason="Unknown windows error in b-casting"
)
def test_create_1d_bcasting_data_2d(s1):
    s2 = s1.copy()
    data = np.random.randint(1, 100, (4, 3))
    for i in range(10):
        j = range(i * 4, i * 4 + 3)
        s2[0, j] = data[0, :]
        s2[1, j] = data[1, :]
        s2[2, j] = data[2, :]
        s2[3, j] = data[3, :]

    s3 = s1.copy()
    for i in range(10):
        j = np.arange(i * 4, i * 4 + 3).reshape(1, -1)
        s3[np.arange(4).reshape(-1, 1), j] = data

    assert s2.spsame(s3)
    assert np.sum(s2 - s3) == 0


def test_create_1d_diag(s1):
    s2 = s1.copy()
    data = np.random.randint(1, 100, len(s1))
    d = np.arange(len(s1))
    for i in d:
        s2[i, i] = data[i]

    s3 = s1.copy()
    s3[d, d] = data

    assert s2.spsame(s3)
    assert np.sum(s2 - s3) == 0


def test_create_2d_diag_0d(s2):
    S1 = s2.copy()
    data = 1
    d = np.arange(len(s2))
    for i in d:
        S1[i, i] = data
    S2 = s2.copy()
    S2[d, d] = data

    S3 = s2.copy()
    S3[d, d, 0] = data
    S3[d, d, 1] = data

    assert S1.spsame(S2)
    assert np.sum(S1 - S2) == 0
    assert S1.spsame(S3)
    assert np.sum(S1 - S3) == 0


def test_create_2d_diag_1d(s2):
    s1 = s2.copy()
    s3 = s2.copy()

    data = np.random.randint(1, 100, len(s1))
    d = np.arange(len(s1))
    for i in d:
        s1[i, i] = data[i]

    s2[d, d] = data
    s3[d, d, 0] = data
    s3[d, d, 1] = data

    assert s1.spsame(s2)
    assert np.sum(s1 - s2) == 0
    assert s1.spsame(s3)
    assert np.sum(s1 - s3) == 0


def test_create_2d_diag_2d(s2):
    s1 = s2.copy()
    s3 = s2.copy()

    data = np.random.randint(1, 100, len(s1) * 2).reshape(-1, 2)
    d = np.arange(len(s1))
    for i in d:
        s1[i, i] = data[i]

    s2[d, d] = data
    s3[d, d, 0] = data[:, 0]
    s3[d, d, 1] = data[:, 1]

    assert s1.spsame(s2)
    assert np.sum(s1 - s2) == 0
    assert s1.spsame(s3)
    assert np.sum(s1 - s3) == 0


def test_create_2d_data_2d(s2):
    s1 = s2.copy()
    s3 = s2.copy()

    # matrix assignment
    I = np.arange(len(s1) // 2)
    data = np.random.randint(1, 100, I.size**2).reshape(I.size, I.size)
    for i in I:
        s1[i, I, 0] = data[i]
        s1[i, I, 1] = data[i]

    I.shape = (-1, 1)

    s2[I, I.T, 0] = data
    s2[I, I.T, 1] = data

    s3[I, I.T] = data[:, :, None]

    assert s1.spsame(s2)
    assert np.sum(s1 - s2) == 0
    assert s1.spsame(s3)
    assert np.sum(s1 - s3) == 0


def test_create_2d_data_3d(s2):
    s1 = s2.copy()

    # matrix assignment
    I = np.arange(len(s1) // 2)
    data = np.random.randint(1, 100, I.size**2 * 2).reshape(I.size, I.size, 2)
    for i in I:
        s1[i, I] = data[i]

    I.shape = (-1, 1)
    s2[I, I.T] = data

    assert s1.spsame(s2)
    assert np.sum(s1 - s2) == 0


def test_copy_dims(s2):
    s2[2, 2] = [2, 3]
    s1 = s2.copy(dims=0)
    assert np.allclose(s2._D[:, 0], s1._D[:, 0])
    s1 = s2.copy(dims=1)
    assert np.allclose(s2._D[:, 1], s1._D[:, 0])

    s1 = s2.copy(dims=[1, 0])
    assert np.allclose(s2._D[:, 1], s1._D[:, 0])
    assert np.allclose(s2._D[:, 0], s1._D[:, 1])


def test_fail_data_3d_to_1d(s2):
    # matrix assignment
    I = np.arange(len(s2) // 2).reshape(-1, 1)
    data = np.random.randint(1, 100, I.size * 2).reshape(I.size, 1, 2)
    with pytest.raises(ValueError):
        s2[I, I.T] = data


def test_fail_data_2d_to_2d(s2):
    # matrix assignment
    I = np.arange(len(s2) // 2).reshape(-1, 1)
    data = np.random.randint(1, 100, I.size**2).reshape(I.size, I.size)
    with pytest.raises(ValueError):
        s2[I, I.T] = data


def test_fail_data_2d_to_3d(s2):
    # matrix assignment
    I = np.arange(len(s2) // 2).reshape(-1, 1)
    data = np.random.randint(1, 100, I.size**2).reshape(I.size, I.size)
    with pytest.raises(ValueError):
        s2[I, I.T, [0, 1]] = data


def test_finalize1(s1):
    s1[0, [1, 2, 3]] = 1
    s1[2, [1, 2, 3]] = 1.0
    s1[1, [3, 2, 1]] = 1.0
    assert not s1.finalized
    p = s1.ptr.view()
    n = s1.ncol.view()
    # Assert that the ordering is good
    assert np.allclose(s1.col[p[1] : p[1] + n[1]], [3, 2, 1])
    s1.finalize()

    assert s1.finalized
    s1.empty(keep_nnz=True)
    assert s1.finalized
    s1.empty()
    assert not s1.finalized


def test_finalize2(s1):
    s1[0, [1, 2, 3]] = 1
    s1[2, [1, 2, 3]] = 1.0
    s1[1, [3, 2, 1]] = 1.0
    assert not s1.finalized
    p = s1.ptr.view()
    n = s1.ncol.view()
    # Assert that the ordering is good
    assert np.allclose(s1.col[p[1] : p[1] + n[1]], [3, 2, 1])
    s1.finalize(False)

    assert not s1.finalized
    assert len(s1.col) == 9


def test_iterator1(s1):
    s1[0, [1, 2, 3]] = 1
    s1[2, [1, 2, 4]] = 1.0
    e = [[1, 2, 3], [], [1, 2, 4]]
    for i, j in s1:
        assert j in e[i]

    for i, j in s1.iter_nnz(0):
        assert i == 0
        assert j in e[0]
    for i, j in s1.iter_nnz(2):
        assert i == 2
        assert j in e[2]

    for i, j in ispmatrix(s1):
        assert j in e[i]

    for i, j in ispmatrix(s1, map_col=lambda x: x):
        assert j in e[i]
    for i, j in ispmatrix(s1, map_row=lambda x: x):
        assert j in e[i]

    for i, j, d in ispmatrixd(s1):
        assert j in e[i]
        assert d == 1.0


def test_iterator2(s1):
    e = [[1, 2, 3], [], [1, 2, 4]]
    s1[0, [1, 2, 3]] = 1
    s1[2, [1, 2, 4]] = 1.0
    a = s1.tocsr()
    for func in ["csr", "csc", "coo", "lil"]:
        a = getattr(a, "to" + func)()
        for r, c in ispmatrix(a):
            assert r in [0, 2]
            assert c in e[r]

    for func in ["csr", "csc", "coo", "lil"]:
        a = getattr(a, "to" + func)()
        for r, c, d in ispmatrixd(a):
            assert r in [0, 2]
            assert c in e[r]
            assert d == 1.0


def test_iterator3(s1):
    e = [[1, 2, 3], [], [1, 2, 4]]
    s1[0, [1, 2, 3]] = 1
    s1[2, [1, 2, 4]] = 1.0
    a = s1.tocsr()
    for func in ["csr", "csc", "coo", "lil"]:
        a = getattr(a, "to" + func)()
        for r, c in ispmatrix(a):
            assert r in [0, 2]
            assert c in e[r]

    # number of mapped values
    nvals = 2
    for func in ["csr", "lil"]:
        a = getattr(a, "to" + func)()
        n = 0
        for r, c in ispmatrix(a, lambda x: x % 2, lambda x: x % 2):
            assert r == 0
            assert c in [0, 1]
            n += 1
        assert n == nvals


def test_delitem_simple(s1):
    s1[0, [1, 2, 3]] = 1
    assert s1.nnz == 3
    del s1[0, 1]
    assert s1.nnz == 2
    assert s1[0, 1] == 0
    assert s1[0, 2] == 1
    assert s1[0, 3] == 1
    s1[0, [1, 2, 3]] = 1
    del s1[0, [1, 3]]
    assert s1.nnz == 1
    assert s1[0, 1] == 0
    assert s1[0, 2] == 1
    assert s1[0, 3] == 0
    del s1[range(2), 0]
    assert s1.nnz == 1
    for i in range(2):
        assert s1[i, 0] == 0


def test_delitem_order():
    s = SparseCSR((10, 10), dtype=np.int32)
    s[0, 3] = 3
    s[0, 2] = 2
    s[0, 1] = 1
    del s[0, 2]
    assert s.nnz == 2
    assert s[0, 1] == 1
    assert s[0, 3] == 3


def test_contains1(s1):
    s1[0, [1, 2, 3]] = 1
    assert s1.nnz == 3
    assert [0, 1] in s1
    assert [0, [1, 3]] in s1


def test_sub1(s1):
    s1[0, [1, 2, 3]] = 1
    assert s1.nnz == 3
    s1 = s1.sub([0, 1])
    assert s1.nnz == 1
    assert s1.shape[0] == 2
    assert s1.shape[1] == 2


def test_remove1(s1):
    s1[0, [1, 2, 3]] = 1
    assert s1.nnz == 3
    s2 = s1.remove([1])
    assert s2.nnz == 2
    assert s2.shape[0] == s1.shape[0] - 1
    assert s2.shape[1] == s1.shape[1] - 1


def test_eliminate_zeros1(s1):
    s1[0, [1, 2, 3]] = 1
    s1[1, [1, 2, 3]] = 0
    assert s1.nnz == 6
    s1.eliminate_zeros()
    assert s1.nnz == 3
    assert s1[1, 1] == 0
    assert s1[1, 2] == 0
    assert s1[1, 3] == 0


def test_eliminate_zeros_tolerance(s1):
    s1[0, [1, 2, 3]] = 1
    s1[1, [1, 2, 3]] = 2
    assert s1.nnz == 6
    s1.eliminate_zeros()
    assert s1.nnz == 6
    s1.eliminate_zeros(1)
    assert s1.nnz == 3
    assert s1[0, 1] == 0
    assert s1[0, 2] == 0
    assert s1[0, 3] == 0


def test_eliminate_zeros_tolerance_ndim():
    s = SparseCSR((3, 3, 3))
    s[1, [0, 1, 2]] = 0.1
    s[1, [0, 1, 2], 1] = 0.05
    assert s.nnz == 3
    s.eliminate_zeros()
    assert s.nnz == 3
    s.eliminate_zeros(0.075)
    assert s.nnz == 3
    assert s[1, 1, 0] == 0.1
    assert s[1, 1, 1] == 0.05
    assert s[1, 1, 2] == 0.1
    s.eliminate_zeros(0.2)
    assert s.nnz == 0


def test_spsame(s1, s2):
    s1[0, [1, 2, 3]] = 1
    s2[0, [1, 2, 3]] = (1, 1)
    assert s1.spsame(s2)
    s2[1, 1] = (1, 1)
    assert not s1.spsame(s2)
    s1.align(s2)
    assert s1.spsame(s2)


@pytest.mark.xfail(reason="same index assignment in single statement")
def test_set_same_index(s1):
    s1[0, [1, 1]] = 1
    assert s1.nnz == 1


def test_delete_col1(s1):
    nc = s1.shape[1]
    s1[0, [1, 3]] = 1
    s1[1, [1, 2, 3]] = 1
    assert s1.nnz == 5
    s1.delete_columns(2)
    assert s1.nnz == 4
    assert s1.shape[1] == nc - 1
    assert np.all(s1.col[array_arange(s1.ptr, n=s1.ncol)] < 3)
    s1.delete_columns(2, True)
    assert s1.nnz == 2
    assert s1.shape[1] == nc - 1
    assert np.all(s1.col[array_arange(s1.ptr, n=s1.ncol)] < 2)
    # Delete a non-existing column
    s1.delete_columns(100000, True)
    assert s1.nnz == 2
    assert s1.shape[1] == nc - 1


def test_delete_col2(s1):
    nc = s1.shape[1]
    s1[1, [1, 2, 3]] = 1
    assert s1.nnz == 3
    s1.delete_columns(3)
    assert s1.nnz == 2
    assert s1.shape[1] == nc - 1
    s1.delete_columns(2, True)
    assert s1.nnz == 1
    assert s1.shape[1] == nc - 1


def test_delete_col3(s1):
    s2 = s1.copy()
    nc = s1.shape[1]
    for i in range(10):
        s1[i, [3, 2, 1]] = 1
        s2[i, 2] = 1
    s1.finalize()
    s2.finalize()
    assert s1.nnz == 10 * 3
    s1.delete_columns([3, 1], keep_shape=True)
    assert s1.ptr[-1] == s1.nnz
    assert s2.ptr[-1] == s2.nnz
    assert s1.nnz == 10 * 1
    assert s1.spsame(s2)


def test_delete_col4():
    s1 = SparseCSR((10, 100), dtype=np.int32)
    s2 = SparseCSR((10, 98), dtype=np.int32)
    nc = s1.shape[1]
    for i in range(10):
        s1[i, [3, 2, 1]] = 1
        s2[i, 1] = 1
    s1.finalize()
    s2.finalize()
    assert s1.nnz == 10 * 3
    s1.delete_columns([3, 1])
    assert s1.ptr[-1] == s1.nnz
    assert s2.ptr[-1] == s2.nnz
    assert s1.nnz == 10 * 1
    assert s1.spsame(s2)


def test_delete_col5(s1):
    nc = s1.shape[1]
    s1[1, [1, 2, 3]] = 1
    assert s1.nnz == 3
    s1.delete_columns(2)
    assert s1.nnz == 2
    assert s1.shape[1] == nc - 1
    assert np.all(s1.col[array_arange(s1.ptr, n=s1.ncol)] < 3)
    s1.delete_columns(2, True)
    assert s1.nnz == 1
    assert s1.shape[1] == nc - 1
    assert np.all(s1.col[array_arange(s1.ptr, n=s1.ncol)] < 2)
    # Delete a non-existing column
    s1.delete_columns(100000, True)
    assert s1.nnz == 1
    assert s1.shape[1] == nc - 1


def test_delete_col6(s1):
    nc = s1.shape[1]
    for i in range(3):
        s1[i, [1, 2, 3]] = 1
    assert s1.nnz == 9
    s1.delete_columns(2)
    assert s1.nnz == 6
    assert s1.shape[1] == nc - 1


def test_translate_col1(s1):
    s1[1, 1] = 1
    s1[1, 2] = 2
    s1[1, 3] = 3
    assert s1.nnz == 3
    s1.translate_columns([1, 3], [3, 1])
    assert s1.nnz == 3
    assert s1[1, 1] == 3
    assert s1[1, 3] == 1


def test_translate_col2(s1):
    s1[1, 1] = 1
    s1[1, 2] = 2
    s1[1, 3] = 3
    assert s1.nnz == 3
    s1.translate_columns([1, 3], [3, s1.shape[1] + 100])
    assert s1.nnz == 2
    assert s1[1, 3] == 1
    assert s1[1, 1] == 0


def test_translate_col3(s1):
    for i in range(3):
        s1[i, 1] = 1
        s1[i, 2] = 2
        s1[i, 3] = 3
    assert s1.nnz == 9
    s1.translate_columns([1, 3], [3, 1])
    assert s1.nnz == 9
    for i in range(3):
        assert s1[i, 1] == 3
        assert s1[i, 3] == 1
    s1.translate_columns([1, 3], [3, 1])
    assert s1.nnz == 9
    for i in range(3):
        assert s1[i, 1] == 1
        assert s1[i, 3] == 3


def test_translate_col4(s1):
    nc = s1.shape[1]
    for i in range(3):
        s1[i, 1] = 1
        s1[i, 2] = 2
        s1[i, 3] = 3
    assert s1.nnz == 9
    s1.translate_columns([1, 3], [nc, 1])
    assert s1.nnz == 6
    assert s1.shape[1] == nc
    for i in range(3):
        assert s1[i, 1] == 3
        assert s1[i, 2] == 2
        assert s1[i, 3] == 0
    s1.translate_columns([1, 3], [3, 1])
    assert s1.nnz == 6
    for i in range(3):
        assert s1[i, 1] == 0
        assert s1[i, 3] == 3


def test_edges1(s1):
    s1[1, 1] = 1
    s1[1, 2] = 2
    s1[1, 3] = 3
    assert np.all(s1.edges(1, exclude=[]) == [1, 2, 3])
    assert np.all(s1.edges(1, exclude=[1]) == [2, 3])
    assert np.all(s1.edges(1, exclude=2) == [1, 3])
    assert len(s1.edges(2)) == 0


def test_nonzero1(s1):
    s1[2, 1] = 1
    s1[1, 1] = 1
    s1[1, 2] = 2
    s1[1, 3] = 3
    assert s1.nnz == 4
    r, c = s1.nonzero()
    assert np.all(r == [1, 1, 1, 2])
    assert np.all(c == [1, 2, 3, 1])
    c = s1.nonzero(only_cols=True)
    assert np.all(c == [1, 2, 3, 1])
    c = s1.nonzero(rows=1, only_cols=True)
    assert np.all(c == [1, 2, 3])
    c = s1.nonzero(rows=2, only_cols=True)
    assert np.all(c == [1])
    c = s1.nonzero(rows=[0, 1], only_cols=True)
    assert np.all(c == [1, 2, 3])
    r, c = s1.nonzero(rows=[0, 1])
    assert np.all(r == [1, 1, 1])
    assert np.all(c == [1, 2, 3])


def test_op1(s1):
    for i in range(10):
        j = range(i * 4, i * 4 + 3)
        s1[0, j] = i

        # i+
        s1 += 1
        for jj in j:
            assert s1[0, jj] == i + 1
            assert s1[1, jj] == 0

        # i-
        s1 -= 1
        for jj in j:
            assert s1[0, jj] == i
            assert s1[1, jj] == 0

        # i*
        s1 *= 2
        for jj in j:
            assert s1[0, jj] == i * 2
            assert s1[1, jj] == 0

        # //
        s1 //= 2
        for jj in j:
            assert s1[0, jj] == i
            assert s1[1, jj] == 0

        # i**
        s1 **= 2
        for jj in j:
            assert s1[0, jj] == i**2
            assert s1[1, jj] == 0


def test_op2(s1):
    for i in range(10):
        j = range(i * 4, i * 4 + 3)
        s1[0, j] = i

        # +
        s = s1 + 1
        for jj in j:
            assert s[0, jj] == i + 1
            assert s1[0, jj] == i
            assert s[1, jj] == 0

        # -
        s = s1 - 1
        for jj in j:
            assert s[0, jj] == i - 1
            assert s1[0, jj] == i
            assert s[1, jj] == 0

        # - (r)
        s = 1 - s1
        for jj in j:
            assert s[0, jj] == 1 - i
            assert s1[0, jj] == i
            assert s[1, jj] == 0

        # *
        s = s1 * 2
        for jj in j:
            assert s[0, jj] == i * 2
            assert s1[0, jj] == i
            assert s[1, jj] == 0

        # *
        s = np.multiply(s1, 2)
        for jj in j:
            assert s[0, jj] == i * 2
            assert s1[0, jj] == i
            assert s[1, jj] == 0
        # *
        s.empty()
        np.multiply(s1, 2, out=s)
        for jj in j:
            assert s[0, jj] == i * 2
            assert s1[0, jj] == i
            assert s[1, jj] == 0

        # //
        s = s // 2
        for jj in j:
            assert s[0, jj] == i
            assert s1[0, jj] == i
            assert s[1, jj] == 0

        # **
        s = s1**2
        for jj in j:
            assert s[0, jj] == i**2
            assert s1[0, jj] == i
            assert s[1, jj] == 0

        # ** (r)
        s = 2**s1
        for jj in j:
            assert s[0, jj], 2 ** s1[0 == jj]
            assert s1[0, jj] == i
            assert s[1, jj] == 0


def test_op_csr(s1):
    csr = sc.sparse.csr_matrix((10, 100), dtype=np.int32)
    for i in range(10):
        j = range(i + 2)
        s1[0, j] = i

        csr[0, 0] = 1

        # +
        s = s1 + csr
        for jj in j:
            if jj == 0:
                continue
            assert s[0, jj] == i
            assert s1[0, jj] == i
            assert s[1, jj] == 0
        assert s[0, 0] == i + 1

        # -
        s = s1 - csr
        for jj in j:
            if jj == 0:
                continue
            assert s[0, jj] == i
            assert s1[0, jj] == i
            assert s[1, jj] == 0
        assert s[0, 0] == i - 1

        # - (r)
        s = csr - s1
        for jj in j:
            if jj == 0:
                continue
            assert s[0, jj] == -i
            assert s1[0, jj] == i
            assert s[1, jj] == 0
        assert s[0, 0] == 1 - i

        csr[0, 0] = 2

        # *
        s = s1 * csr
        for jj in j:
            if jj == 0:
                continue
            assert s[0, jj] == 0
            assert s1[0, jj] == i
            assert s[1, jj] == 0
        assert s[0, 0] == i * 2

        # //
        s = s // csr
        assert s[0, 0] == i

        # **
        s = s1**csr
        for jj in j:
            if jj == 0:
                continue
            assert s[0, jj] == 1
            assert s1[0, jj] == i
            assert s[1, jj] == 0
        assert s[0, 0] == i**2


def test_op3():
    S = SparseCSR((10, 100), dtype=np.int32)
    # Create initial stuff
    for i in range(10):
        j = range(i * 4, i * 4 + 3)
        S[0, j] = i

    for op in ("add", "sub", "mul", "pow"):
        func = getattr(S, f"__{op}__")
        s = func(1)
        assert s.dtype == np.int32
        s = func(1.0)
        assert s.dtype == np.float64
        if op != "pow":
            s = func(1.0j)
            assert s.dtype == np.complex128

    S = S.copy(dtype=np.float64)
    for op in ("add", "sub", "mul", "pow"):
        func = getattr(S, f"__{op}__")
        s = func(1)
        assert s.dtype == np.float64
        s = func(1.0)
        assert s.dtype == np.float64
        if op != "pow":
            s = func(1.0j)
            assert s.dtype == np.complex128

    S = S.copy(dtype=np.complex128)
    for op in ("add", "sub", "mul", "pow"):
        func = getattr(S, f"__{op}__")
        s = func(1)
        assert s.dtype == np.complex128
        s = func(1.0)
        assert s.dtype == np.complex128
        if op != "pow":
            s = func(1.0j)
            assert s.dtype == np.complex128


def test_op4():
    S = SparseCSR((10, 100), dtype=np.int32)
    # Create initial stuff
    for i in range(10):
        j = range(i * 4, i * 4 + 3)
        S[0, j] = i

    s = 1 + S
    assert s.dtype == np.int32
    s = 1.0 + S
    assert s.dtype == np.float64
    s = 1.0j + S
    assert s.dtype == np.complex128

    s = 1 - S
    assert s.dtype == np.int32
    s = 1.0 - S
    assert s.dtype == np.float64
    s = 1.0j - S
    assert s.dtype == np.complex128

    s = 1 * S
    assert s.dtype == np.int32
    s = 1.0 * S
    assert s.dtype == np.float64
    s = 1.0j * S
    assert s.dtype == np.complex128

    s = 1**S
    assert s.dtype == np.int32
    s = 1.0**S
    assert s.dtype == np.float64
    s = 1.0j**S
    assert s.dtype == np.complex128


def binary():
    op = operator
    ops = [
        op.mod,
        op.mul,
        op.add,
        op.sub,
        op.floordiv,
        op.truediv,
    ]
    return ops


def binary_int():
    op = operator
    return [op.pow]


def unary():
    op = operator
    ops = [
        op.abs,
        op.neg,
        op.pos,
    ]
    return ops


@pytest.fixture(scope="module")
def matrix_sisl_csr():
    matrix = []

    def add3(m):
        matrix.append(m)
        m = m.copy()
        m.finalize(sort=False)
        matrix.append(m)
        m = m.copy()
        m.finalize(sort=True)
        matrix.append(m)

    # diagonal
    m = SparseCSR((10, 80), dtype=np.int32)
    for i in range(10):
        m[i, i] = 1
    add3(m)

    # not completely full (some empty ncol)

    m = SparseCSR((10, 80), dtype=np.int32)
    m[0, [1, 0]] = [12, 2]
    m[2, 2] = 1
    m[4, [4, 0]] = [2, 3]
    m[5, [5, 0]] = [3, 5]
    add3(m)

    # all more than 1 coupling, and not sorted
    m = SparseCSR((10, 80), dtype=np.int32)
    m[0, [1, 0]] = [11, 0]
    m[1, [10, 50, 20]] = [14, 20, 43]
    m[2, [4, 7, 3, 1]] = [2, 5, 4, 10]
    m[3, [4, 7, 3, 1]] = [2, 5, 4, 10]
    m[4, [1, 5, 3, 21]] = [2, 5, 4, 10]
    m[5, [53, 52, 3, 21]] = [31, 6, 7, 12]
    m[6, [43, 32, 1, 6]] = [3, 6, 7, 16]
    m[7, [65, 44, 3, 2]] = [3, 6, 73, 31]
    m[8, [66, 45, 6, 8]] = [4, 3, 12, 357]
    m[9, [55, 44, 33, 22]] = [4, 3, 11, 27]
    add3(m)
    return matrix


@pytest.fixture(scope="module")
def matrix_csr_matrix():
    matrix = []

    csr_matrix = sc.sparse.csr_matrix

    def add3(m):
        matrix.append(m)
        m = m.copy()
        m.sort_indices()
        matrix.append(m)

    # diagonal
    m = csr_matrix((10, 80), dtype=np.int32)
    for i in range(10):
        m[i, i] = 1
    add3(m)

    # not completely full (some empty ncol)

    m = csr_matrix((10, 80), dtype=np.int32)
    m[0, [1, 0]] = [11, 3]
    m[2, 2] = 10
    m[4, [4, 0]] = [22, 40]
    m[5, [5, 0]] = [33, 51]
    add3(m)

    # all more than 1 coupling, and not sorted
    m = csr_matrix((10, 80), dtype=np.int32)
    m[0, [1, 0]] = [12, 4]
    m[1, [10, 50, 20]] = [11, 20, 401]
    m[2, [4, 7, 3, 1]] = [2, 5, 4, 10]
    m[3, [4, 7, 3, 1]] = [2, 5, 4, 10]
    m[4, [1, 5, 3, 21]] = [2, 5, 4, 10]
    m[5, [53, 52, 3, 21]] = [3, 6, 7, 14]
    m[6, [43, 32, 1, 6]] = [3, 6, 7, 11]
    m[7, [65, 44, 3, 2]] = [3, 6, 7, 12]
    m[8, [66, 45, 6, 8]] = [4, 3, 15, 7]
    m[9, [55, 44, 33, 22]] = [4, 3, 14, 7]
    add3(m)
    return matrix


@pytest.mark.parametrize("op", binary())
def test_op_binary(matrix_sisl_csr, matrix_csr_matrix, op):
    for m in matrix_sisl_csr:
        mD = m.toarray()[..., 0]
        if op not in (operator.add, operator.sub):
            v = op(m, 3)
            assert np.allclose(op(mD, 3), v.toarray()[..., 0])

        if op not in (operator.truediv, operator.floordiv):
            for m2 in matrix_sisl_csr:
                m2D = m2.toarray()[..., 0]
                v = op(m, m2)
                assert np.allclose(op(mD, m2D), v.toarray()[..., 0])

            for m2 in matrix_csr_matrix:
                m2D = m2.toarray()
                v = op(m, m2)
                assert np.allclose(op(mD, m2D), v.toarray()[..., 0])


def test_op5():
    S1 = SparseCSR((10, 100), dtype=np.int32)
    S2 = SparseCSR((10, 100), dtype=np.int32)
    S3 = SparseCSR((10, 100), dtype=np.int32)
    # Create initial stuff
    for i in range(10):
        j = range(i * 4, i * 4 + 3)
        S1[0, j] = i
        S2[0, j] = i

        if i < 5:
            S3[0, j] = i

    S = S1 * S2
    assert np.allclose(S.todense(), (S2**2).todense())

    S = S * S
    assert np.allclose(S.todense(), (S2**4).todense())

    S = S1 + S2
    S1 += S2
    assert np.allclose(S.todense(), S1.todense())

    S = S1 - S2
    S1 -= S2
    assert np.allclose(S.todense(), S1.todense())

    S = S1 + 2
    S -= 2
    assert np.allclose(S.todense(), S1.todense())

    S = S1 * 2
    S //= 2
    assert np.allclose(S.todense(), S1.todense())

    S = S1 / 2.0
    S *= 2
    assert np.allclose(S.todense(), S1.todense())


def test_op_numpy_scalar():
    S = SparseCSR((10, 100), dtype=np.float32)
    I = np.ones(1, dtype=np.complex64)[0]
    # Create initial stuff
    for i in range(10):
        j = range(i * 4, i * 4 + 3)
        S[0, j] = i
    S.finalize()

    Ssum = S._D.sum()

    s = S + I
    assert isinstance(s, SparseCSR)
    assert s.dtype == np.complex64
    assert s._D.sum() == Ssum + S.nnz

    s = S - I
    assert isinstance(s, SparseCSR)
    assert s.dtype == np.complex64
    assert s._D.sum() == Ssum - S.nnz

    s = I + S
    assert isinstance(s, SparseCSR)
    assert s.dtype == np.complex64
    assert s._D.sum() == Ssum + S.nnz

    s = S * I
    assert isinstance(s, SparseCSR)
    assert s.dtype == np.complex64
    assert s._D.sum() == Ssum

    s = I * S
    assert isinstance(s, SparseCSR)
    assert s.dtype == np.complex64
    assert s._D.sum() == Ssum

    s = S / I
    assert isinstance(s, SparseCSR)
    assert s.dtype == np.complex64
    assert s._D.sum() == Ssum

    s = S**I
    assert isinstance(s, SparseCSR)
    assert s.dtype == np.complex64
    assert s._D.sum() == Ssum

    s = I**S
    assert isinstance(s, SparseCSR)
    assert s.dtype == np.complex64

    s = 1j * S
    assert isinstance(s, SparseCSR)
    assert s.dtype == (1j * np.array([1j], dtype=np.complex64)).dtype

    s = np.exp(s)
    assert isinstance(s, SparseCSR)
    assert s.dtype == np.exp(1j * np.array([1j], dtype=np.complex64)).dtype


def test_op_sparse_dim():
    S = SparseCSR((10, 100, 2), dtype=np.float32)
    assert S.shape == (10, 100, 2)
    I = np.ones(1, dtype=np.complex64)[0]
    # Create initial stuff
    for i in range(10):
        j = range(i * 4, i * 4 + 3)
        S[0, j] = i
    S.finalize()

    # Try and add different values to the last 2 dimensions
    s = S + [2, 2]
    assert np.allclose(s._D, (S + 2)._D)
    s = S + [1, 2]
    assert np.allclose(s._D[:, 0], (S + 1)._D[:, 0])

    # check both
    ss = S + 1
    assert np.any(np.isclose(s._D[:, 0], ss._D[:, 0]))
    assert not np.any(np.isclose(s._D[:, 1], ss._D[:, 1]))

    ss = S + 2
    assert not np.any(np.isclose(s._D[:, 0], ss._D[:, 0]))
    assert np.all(np.isclose(s._D[:, 1], ss._D[:, 1]))


def test_sparse_transpose():
    S = SparseCSR((10, 100, 2), dtype=np.float32)
    assert S.shape == (10, 100, 2)
    I = np.ones(1, dtype=np.complex64)[0]
    # Create initial stuff
    for i in range(10):
        j = range(i * 4, i * 4 + 3)
        S[0, j] = i
    S.finalize()

    s = S.transpose(False)
    assert s.shape == (100, 10, 2)

    s = s.transpose(False)
    assert s.shape == (10, 100, 2)


def test_op_reduce():
    S1 = SparseCSR((10, 11, 2), dtype=np.int32)
    S1[0, 0] = [1, 2]
    S1[2, 0] = [1, 2]
    S1[2, 2] = [1, 2]

    S2 = np.sum(S1, axis=-1)
    assert S1.spsame(S2)
    assert S2[0, 0] == 3
    assert S2[0, 1] == 0
    assert S2[2, 0] == 3
    assert S2[2, 2] == 3

    assert np.sum(S1) == 1 * 3 + 2 * 3

    S = np.sum(S1, axis=0)
    v = np.zeros([S1.shape[1], S1.shape[2]], np.int32)
    v[0] = [2, 4]
    v[2] = [1, 2]
    assert np.allclose(S, v)

    v = v.sum(1)
    assert np.allclose(np.sum(S, 1), v)

    S = np.sum(S1, axis=1)
    v = np.zeros([S1.shape[0], S1.shape[2]], np.int32)
    v[0] = [1, 2]
    v[2] = [2, 4]
    assert np.allclose(S, v)


def test_unfinalized_math():
    S1 = SparseCSR((4, 4, 1))
    S2 = SparseCSR((4, 4, 1))
    S1[0, 0] = 2.0
    S1[1, 2] = 3.0
    S2[2, 3] = 4.0
    S2[2, 2] = 3.0
    S2[0, 0] = 4

    for i in range(3):
        assert np.allclose(S1.todense() + S2.todense(), (S1 + S2).todense())
        assert np.allclose(S1.todense() * S2.todense(), (S1 * S2).todense())
        sin = np.sin(S1.todense()) + np.sin(S2.todense())
        sins = (np.sin(S1) + np.sin(S2)).todense()
        assert np.allclose(sin, sins)
        sins = np.sin(S1).todense() + np.sin(S2).todense()
        assert np.allclose(sin, sins)

        if i == 0:
            assert not S1.finalized
            assert not S2.finalized
            S1.finalize()
        elif i == 1:
            assert S1.finalized
            assert not S2.finalized
            S2.finalize()
        elif i == 2:
            assert S1.finalized
            assert S2.finalized


def test_pickle():
    import pickle as p

    S = SparseCSR((10, 10, 2), dtype=np.int32)
    S[0, 0] = [1, 2]
    S[2, 0] = [1, 2]
    S[2, 2] = [1, 2]
    n = p.dumps(S)
    s = p.loads(n)
    assert s.spsame(S)


@pytest.mark.parametrize("i", [-1, 10])
def test_sparse_row_out_of_bounds(i):
    S = SparseCSR((10, 10, 1), dtype=np.int32)
    with pytest.raises(IndexError):
        S[i, 0] = 1


@pytest.mark.parametrize("j", [-1, 10])
def test_sparse_column_out_of_bounds(j):
    S = SparseCSR((10, 10, 1), dtype=np.int32)
    with pytest.raises(IndexError):
        S[0, j] = 1


def test_fromsp_csr():
    csr1 = sc.sparse.random(10, 100, 0.01, random_state=24812)
    csr2 = sc.sparse.random(10, 100, 0.02, random_state=24813)

    csr = SparseCSR.fromsp([csr1, csr2])
    csr_1 = csr.tocsr(0)
    csr_2 = csr.tocsr(1)

    assert np.abs(csr1 - csr_1).sum() == 0.0
    assert np.abs(csr2 - csr_2).sum() == 0.0


def test_transform1():
    csr1 = sc.sparse.random(10, 100, 0.01, random_state=24812)
    csr2 = sc.sparse.random(10, 100, 0.02, random_state=24813)
    csr = SparseCSR.fromsp([csr1, csr2])

    # real 1x2 matrix, dtype=np.complex128
    matrix = [[0.3, 0.7]]
    tr = csr.transform(matrix=matrix, dtype=np.complex128)

    assert tr.shape[:2] == csr.shape[:2]
    assert tr.shape[2] == len(matrix)
    assert np.abs(tr.tocsr(0) - 0.3 * csr1 - 0.7 * csr2).sum() == 0.0


def test_transform2():
    csr1 = sc.sparse.random(10, 100, 0.01, random_state=24812)
    csr2 = sc.sparse.random(10, 100, 0.02, random_state=24813)
    csr = SparseCSR.fromsp([csr1, csr2])

    # real 2x2 matrix, dtype=np.float64
    matrix = [[0.3, 0], [0, 0.7]]
    tr = csr.transform(matrix=matrix, dtype=np.float64)

    assert tr.shape[:2] == csr.shape[:2]
    assert tr.shape[2] == len(matrix)
    assert np.abs(tr.tocsr(0) - 0.3 * csr1).sum() == 0.0
    assert np.abs(tr.tocsr(1) - 0.7 * csr2).sum() == 0.0


def test_transform3():
    csr1 = sc.sparse.random(10, 100, 0.01, random_state=24812)
    csr2 = sc.sparse.random(10, 100, 0.02, random_state=24813)
    csr = SparseCSR.fromsp([csr1, csr2])

    # real 3x2 matrix
    matrix = [[0.3, 0], [0, 0.7], [0.1, 0.2]]
    tr = csr.transform(matrix=matrix)

    assert tr.shape[:2] == csr.shape[:2]
    assert tr.shape[2] == len(matrix)
    assert np.abs(tr.tocsr(0) - 0.3 * csr1).sum() == 0.0
    assert np.abs(tr.tocsr(1) - 0.7 * csr2).sum() == 0.0
    assert np.abs(tr.tocsr(2) - 0.1 * csr1 - 0.2 * csr2).sum() == 0.0


def test_transform4():
    csr1 = sc.sparse.random(10, 100, 0.01, random_state=24812)
    csr2 = sc.sparse.random(10, 100, 0.02, random_state=24813)
    csr = SparseCSR.fromsp([csr1, csr2])

    # complex 1x2 matrix
    matrix = [[0.3j, 0.7j]]
    tr = csr.transform(matrix=matrix)

    assert tr.shape[:2] == csr.shape[:2]
    assert tr.shape[2] == len(matrix)
    assert np.abs(tr.tocsr(0) - 0.3j * csr1 - 0.7j * csr2).sum() == 0.0


def test_transform_fail():
    csr1 = sc.sparse.random(10, 100, 0.01, random_state=24812)
    csr2 = sc.sparse.random(10, 100, 0.02, random_state=24813)
    csr = SparseCSR.fromsp((csr1, csr2))

    # complex 1x3 matrix
    matrix = [[0.3j, 0.7j, 1.0]]
    with pytest.raises(ValueError):
        csr.transform(matrix=matrix)


@pytest.mark.slow
def test_fromsp_csr_large():
    csr1 = sc.sparse.random(10000, 10, 0.1, format="csr", random_state=23583)
    csr2 = csr1.copy()

    print_time = False

    from time import time

    # Add some more stuff
    row = csr1[9948]
    indices = row.indices
    if len(indices) == 0:
        indices = np.arange(3)
    csr2[9948, (indices + 1) % 10] = 1.0
    assert csr1.getnnz() != csr2.getnnz()

    t0 = time()
    csr = SparseCSR.fromsp([csr1, csr2])
    if print_time:
        print(f"timing: fromsp {time() - t0}")
    csr_1 = csr.tocsr(0)
    csr_2 = csr.tocsr(1)

    assert np.abs(csr1 - csr_1).sum() == 0.0
    assert np.abs(csr2 - csr_2).sum() == 0.0

    csr_ = SparseCSR(csr1.shape + (2,), nnzpr=1)

    t0 = time()
    for ic, c in enumerate([csr1, csr2]):
        ptr = c.indptr

        # Loop stuff
        for r in range(c.shape[0]):
            idx = csr_._extend(r, c.indices[ptr[r] : ptr[r + 1]])
            csr_._D[idx, ic] += c.data[ptr[r] : ptr[r + 1]]
    if print_time:
        print(f"timing: 2 x ptr[]:ptr[] {time() - t0}")

    dcsr = csr - csr_

    assert np.abs(dcsr.tocsr(0)).sum() == 0.0
    assert np.abs(dcsr.tocsr(1)).sum() == 0.0

    csr_ = SparseCSR(csr1.shape + (2,), nnzpr=1)

    t0 = time()
    for ic, c in enumerate([csr1, csr2]):
        ptr = c.indptr

        # Loop stuff
        for r in range(c.shape[0]):
            sl = slice(ptr[r], ptr[r + 1])
            idx = csr_._extend(r, c.indices[sl])
            csr_._D[idx, ic] += c.data[sl]
    if print_time:
        print(f"timing: slice(ptr[]:ptr[]) {time() - t0}")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import pytest

from sisl.mixing import History

pytestmark = pytest.mark.mixing


def test_simple():
    hist = History(5)

    assert len(hist) == 0
    assert hist.max_elements == 5

    hist.append(1, 2)
    assert len(hist) == 1
    assert len(hist[0]) == 2
    hist.append(1, 2, 3)
    assert len(hist) == 2
    assert len(hist[0]) == 2
    assert len(hist[1]) == 3
    hist.append(1, 2, 3)
    assert len(hist) == 3
    assert len(hist[0]) == 2
    assert len(hist[1]) == 3
    assert len(hist[2]) == 3

    # test clear
    hist.clear(0)
    assert len(hist) == 2
    hist.append(1)
    assert len(hist) == 3
    hist.clear()
    assert len(hist) == 0



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import platform
import sys

import numpy as np
import pytest
import scipy.sparse as sps

from sisl import Atom, SislWarning, SparseCSR, Spin, geom
from sisl.physics.sparse import SparseOrbitalBZ, SparseOrbitalBZSpin

pytestmark = [pytest.mark.physics, pytest.mark.sparse]


def _get():
    gr = geom.graphene()
    return gr


def test_str():
    gr = _get()
    # The most simple setup.
    sp = SparseOrbitalBZ(gr)
    str(sp)


def run_Pk_hermitian_tests(M):
    for k in ([0, 0, 0], [0.1, 0.2, 0.3]):
        Mk = M.Pk(k=k, format="array")
        assert np.allclose(Mk, Mk.T.conj())
        Mk = M.Pk(k=k).toarray()
        assert np.allclose(Mk, Mk.T.conj())


def test_S():
    gr = _get()
    # The most simple setup.
    sp = SparseOrbitalBZ(gr, orthogonal=False)
    sp[0, 0] = 0.5
    sp[1, 1] = 0.5
    sp.S[0, 0] = 1.0
    sp.S[1, 1] = 1.0
    assert sp[1, 1, 0] == pytest.approx(0.5)
    assert sp.S[1, 1] == pytest.approx(1.0)


def test_eigh_orthogonal():
    gr = _get()
    # The most simple setup.
    sp = SparseOrbitalBZ(gr)
    sp[0, 0] = 0.5
    sp[1, 1] = 0.5
    assert np.allclose(sp.eigh(), [0.5, 0.5])
    assert np.allclose(sp.eigh([0.5] * 3), [0.5, 0.5])


def test_eigh_non_orthogonal():
    gr = _get()
    # The most simple setup.
    sp = SparseOrbitalBZ(gr, orthogonal=False)
    sp[0, 0] = 0.5
    sp[1, 1] = 0.5
    sp.S[0, 0] = 1.0
    sp.S[1, 1] = 1.0
    assert np.allclose(sp.eigh(), [0.5, 0.5])


def test_eigsh_orthogonal():
    gr = _get()
    # The most simple setup.
    sp = SparseOrbitalBZ(gr)
    sp[0, 0] = 0.5
    sp[1, 1] = 0.5
    # Fails due to too many requested eigenvalues
    with pytest.raises(TypeError):
        sp.eigsh(n=3)


def test_eigsh_non_orthogonal():
    sp = SparseOrbitalBZ(_get(), orthogonal=False)
    sp.construct([(0.1, 1.44), ([0, 1.0], [-2.7, 0])])
    sp.eigsh(n=1)


def test_pickle_non_orthogonal():
    import pickle as p

    gr = _get()
    sp = SparseOrbitalBZ(gr, orthogonal=False)
    sp[0, 0] = 0.5
    sp[1, 1] = 0.5
    sp.S[0, 0] = 1.0
    sp.S[1, 1] = 1.0
    s = p.dumps(sp)
    SP = p.loads(s)
    assert sp.spsame(SP)
    assert np.allclose(sp.eigh(), SP.eigh())


def test_pickle_non_orthogonal_spin():
    import pickle as p

    gr = _get()
    sp = SparseOrbitalBZSpin(gr, spin=Spin("p"), orthogonal=False)
    sp[0, 0, :] = 0.5
    sp[1, 1, :] = 0.5
    sp.S[0, 0] = 1.0
    sp.S[1, 1] = 1.0
    s = p.dumps(sp)
    SP = p.loads(s)
    assert sp.spsame(SP)
    assert np.allclose(sp.eigh(), SP.eigh())


@pytest.mark.parametrize("n0", [1, 2])
@pytest.mark.parametrize("n1", [1, 3])
@pytest.mark.parametrize("n2", [1, 4])
def test_sparse_orbital_bz_hermitian(sisl_allclose, n0, n1, n2):
    g = geom.fcc(1.0, Atom(1, R=1.5)) * 2
    s = SparseOrbitalBZ(g)
    s.construct([[0.1, 1.51], [1, 2]])
    s = s.tile(n0, 0).tile(n1, 1).tile(n2, 2)
    no = s.geometry.no

    nnz = 0
    for io in range(no):
        # orbitals connecting to io
        edges = s.edges(io)
        # Figure out the transposed supercell indices of the edges
        isc = -s.geometry.o2isc(edges)
        # Convert to supercell
        IO = s.geometry.lattice.sc_index(isc) * no + io
        # Figure out if 'io' is also in the back-edges
        for jo, edge in zip(IO, edges % no):
            assert jo in s.edges(edge)
            nnz += 1

    # Check that we have counted all nnz
    assert s.nnz == nnz

    # Since we are also dealing with f32 data-types we cannot go beyond 1e-7
    for k0 in [0, 0.1]:
        for k1 in [0, -0.15]:
            for k2 in [0, 0.33333]:
                k = (k0, k1, k2)

                if np.allclose(k, 0.0):
                    dtypes = [None, np.float32, np.float64]
                else:
                    dtypes = [None, np.complex64, np.complex128]

                # Also assert Pk == Pk.H for all data-types
                for dtype in dtypes:
                    allclose = sisl_allclose[dtype]

                    Pk = s.Pk(k=k, format="csr", dtype=dtype)
                    assert allclose(Pk.toarray(), Pk.getH().toarray())

                    Pk = s.Pk(k=k, format="array", dtype=dtype)
                    assert allclose(Pk, Pk.T.conj())


def test_sparse_orbital_bz_non_colinear(sisl_allclose):
    M = SparseOrbitalBZSpin(geom.graphene(), spin=Spin("NC"))
    M.construct(([0.1, 1.44], [[0.1, 0.2, 0.3, 0.4], [0.2, 0.3, 0.4, 0.5]]))
    M.finalize()

    MT = M.transpose(spin=True)
    MH = M.transpose(conjugate=True, spin=True)

    allclose = sisl_allclose[M.dtype]

    assert not allclose((M - MT)._csr._D, 0)
    # For a non-collinear with construct we don't take
    # into account the imaginary parts... :(
    # Transposing and Hermitian transpose are the same for NC
    # There are only 1 imaginary part which will change sign regardless
    assert not allclose((MT - MH)._csr._D, 0)
    assert allclose((M - MH)._csr._D, 0)
    run_Pk_hermitian_tests(M)
    run_Pk_hermitian_tests(MH)


def _so_real2cmplx(p):
    return [p[0] + 1j * p[4], p[1] + 1j * p[5], p[2] + 1j * p[3], p[6] + 1j * p[7]]


@pytest.mark.parametrize("dtype", [np.float64, np.complex128])
def test_sparse_orbital_bz_spin_orbit_warns_hermitian(dtype):
    M = SparseOrbitalBZSpin(geom.graphene(), spin=Spin("SO"), dtype=dtype)

    p0 = np.arange(1, 9) / 10
    p1 = np.arange(2, 10) / 10

    if dtype == np.complex128:
        p0 = _so_real2cmplx(p0)
        p1 = _so_real2cmplx(p1)

    with pytest.warns(SislWarning, match="Hermitian"):
        M.construct(([0.1, 1.44], [p0, p1]))


@pytest.mark.parametrize("dtype", [np.float64, np.complex128])
def test_sparse_orbital_bz_spin_orbit(dtype):
    M = SparseOrbitalBZSpin(geom.graphene(), spin=Spin("SO"), dtype=dtype)

    p0 = [0.1, 0.2, 0.3, 0.4, 0.0, 0.0, 0.3, -0.4]
    p1 = [0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9]

    if dtype == np.complex128:
        p0 = _so_real2cmplx(p0)
        p1 = _so_real2cmplx(p1)

    M.construct(
        (
            [0.1, 1.44],
            [p0, p1],
        )
    )
    M.finalize()

    MT = M.transpose(spin=True)
    MH = M.transpose(conjugate=True, spin=True)

    assert np.abs((M - MT)._csr._D).sum() != 0
    assert np.abs((M - MH)._csr._D).sum() == 0
    assert np.abs((MT - MH)._csr._D).sum() != 0
    run_Pk_hermitian_tests(M)
    run_Pk_hermitian_tests(MH)


def test_sparse_orbital_bz_spin_orbit_astype():
    Mr = SparseOrbitalBZSpin(geom.graphene(), spin=Spin("SO"), dtype=np.float64)
    Mc = SparseOrbitalBZSpin(geom.graphene(), spin=Spin("SO"), dtype=np.complex128)

    p0 = [0.1, 0.2, 0.3, 0.4, 0.0, 0.0, 0.3, -0.4]
    p1 = [0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9]

    Mr.construct(
        (
            [0.1, 1.44],
            [p0, p1],
        )
    )
    p0 = _so_real2cmplx(p0)
    p1 = _so_real2cmplx(p1)
    Mc.construct(
        (
            [0.1, 1.44],
            [p0, p1],
        )
    )

    assert np.allclose(Mr.astype(np.complex128)._csr._D, Mc._csr._D)
    assert np.allclose(Mc.astype(np.float64)._csr._D, Mr._csr._D)
    assert np.allclose(Mr.astype(np.complex128).astype(np.float64)._csr._D, Mr._csr._D)
    assert np.allclose(Mc.astype(np.float64).astype(np.complex128)._csr._D, Mc._csr._D)


def _nambu_cmplx2real(p):
    return [
        p[0].real,
        p[1].real,
        p[2].real,
        p[2].imag,
        p[0].imag,
        p[1].imag,
        p[3].real,
        p[3].imag,
        p[4].real,
        p[4].imag,
        p[5].real,
        p[5].imag,
        p[6].real,
        p[6].imag,
        p[7].real,
        p[7].imag,
    ]


@pytest.mark.filterwarnings("ignore", message="*non-Hermitian on-site")
def test_sparse_orbital_bz_nambu_astype():
    Mr = SparseOrbitalBZSpin(geom.graphene(), spin=Spin("nambu"), dtype=np.float64)
    Mc = SparseOrbitalBZSpin(geom.graphene(), spin=Spin("nambu"), dtype=np.complex128)

    p0 = [
        0.1 + 1j * 0.0,
        0.2 + 1j * 0.0,
        0.3 + 1j * 0.4,
        0.3 - 1j * 0.4,
        # onsite S must have zero real
        # onsite triplet states must have 0 imaginary
        1j * 0.6,
        0.3,
        0.4,
        0.3,
    ]

    p1 = [
        0.2 + 1j * 0.6,
        0.3 + 1j * 0.7,
        0.4 + 1j * 0.5,
        0.3 + 1j * 0.9,
        0.3 + 1j * 0.7,
        0.4 + 1j * 0.8,
        0.5 + 1j * 0.6,
        0.4 + 1j * 1.0,
    ]

    Mc.construct(
        (
            [0.1, 1.44],
            [p0, p1],
        )
    )
    p0 = _nambu_cmplx2real(p0)
    p1 = _nambu_cmplx2real(p1)
    Mr.construct(
        (
            [0.1, 1.44],
            [p0, p1],
        )
    )

    assert np.allclose(Mr.astype(np.complex128)._csr._D, Mc._csr._D)
    assert np.allclose(Mc.astype(np.float64)._csr._D, Mr._csr._D)
    assert np.allclose(Mr.astype(np.complex128).astype(np.float64)._csr._D, Mr._csr._D)
    assert np.allclose(Mc.astype(np.float64).astype(np.complex128)._csr._D, Mc._csr._D)


@pytest.mark.parametrize("dtype", [np.float64, np.complex128])
def test_sparse_orbital_bz_spin_orbit_hermitian_not(dtype):
    M = SparseOrbitalBZSpin(geom.graphene(), spin="SO", dtype=dtype)

    p0 = [0.1, 0.2, 0.3, 0.4, 0.0, 0.0, 0.3, -0.4]
    p1 = [0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9]

    if dtype == np.complex128:
        p0 = _so_real2cmplx(p0)
        p1 = _so_real2cmplx(p1)

    M.construct(
        (
            [0.1, 1.44],
            [p0, p1],
        )
    )
    M.finalize()
    new = (M + M.transpose(conjugate=True, spin=True)) / 2
    assert np.abs((M - new)._csr._D).sum() == 0


@pytest.mark.filterwarnings("ignore", message="*non-Hermitian on-site")
@pytest.mark.parametrize(
    "spin", ["unpolarized", "polarized", "non-colinear", "spin-orbit", "nambu"]
)
@pytest.mark.parametrize("finalize", [True, False])
@pytest.mark.parametrize("dtype", [np.float64, np.complex128])
def test_sparse_orbital_spin_make_hermitian(spin, finalize, dtype, sisl_allclose):
    M = SparseOrbitalBZSpin(
        geom.graphene(), spin=spin, dtype=np.complex128, orthogonal=False
    )
    ns = M.shape[-1]

    p0 = np.random.rand(ns) + 1j * np.random.rand(ns)
    p1 = np.random.rand(ns) + 1j * np.random.rand(ns)

    M.construct(
        (
            [0.1, 1.44],
            [p0, p1],
        )
    )
    if finalize:
        M.finalize()
    # We have to do it after to ensure we correctly populate NC+Nambu
    M = M.astype(dtype)

    allclose = sisl_allclose[M.dtype]

    MH = (M + M.transpose(conjugate=True, spin=True)) / 2
    assert allclose((MH - MH.transpose(conjugate=True, spin=True))._csr._D, 0)

    for format, proc in (("array", lambda x: x), ("csr", lambda x: x.toarray())):
        for k in ([0, 0, 0], [0.1, 0.2, 0.3]):
            Mk = proc(MH.Pk(k=k, format=format))
            assert allclose(Mk, Mk.T.conj())
            Mk = proc(MH.Sk(k=k, format=format))
            assert allclose(Mk, Mk.T.conj())


@pytest.mark.filterwarnings("ignore", message="*non-Hermitian on-site")
@pytest.mark.parametrize(
    "spin", ["unpolarized", "polarized", "non-colinear", "spin-orbit", "nambu"]
)
@pytest.mark.parametrize("dtype", [np.float64, np.complex128])
def test_sparse_orbital_spin_make_trs(spin, dtype, sisl_allclose):
    M = SparseOrbitalBZSpin(
        geom.graphene(), spin=spin, dtype=np.complex128, orthogonal=False
    )
    ns = M.shape[-1]

    p0 = np.random.rand(ns) + 1j * np.random.rand(ns)
    p1 = np.random.rand(ns) + 1j * np.random.rand(ns)
    # overlap matrices needs to have certain structure, otherwise
    # diagonalization will fail!
    p0[-1] = 1 + 1j * p0[-1].imag
    p1[-1] = 0.2 + 1j * p1[-1].imag

    M.construct(
        (
            [0.1, 1.44],
            [p0, p1],
        )
    )
    M.finalize()
    M = M.astype(dtype)

    allclose = sisl_allclose[M.dtype]

    MTRS = M.trs()
    if spin == "unpolarized":
        if np.dtype(dtype).kind == "c":
            assert not allclose((M - MTRS)._csr._D, 0)
    else:
        assert not allclose((M - MTRS)._csr._D, 0)
    MTRS = (M + MTRS) / 2
    assert allclose((MTRS - MTRS.trs())._csr._D, 0)

    k = np.array([0.1, 0.2, 0.3])
    eig1 = MTRS.eigh(k=k)
    eig2 = MTRS.eigh(k=-k)
    assert allclose(eig1, eig2)


@pytest.mark.filterwarnings("ignore", message="*non-Hermitian on-site")
@pytest.mark.parametrize(
    "spin", ["unpolarized", "polarized", "non-colinear", "spin-orbit", "nambu"]
)
@pytest.mark.parametrize("dtype", [np.float64, np.complex128])
@pytest.mark.parametrize("orthogonal", [True, False])
def test_sparse_orbital_spin_transpose(spin, dtype, orthogonal, sisl_allclose):
    M = SparseOrbitalBZSpin(
        geom.graphene(), spin=spin, orthogonal=orthogonal, dtype=np.complex128
    )
    ns = M.shape[-1]

    p0 = np.random.rand(ns) + 1j * np.random.rand(ns)
    p1 = np.random.rand(ns) + 1j * np.random.rand(ns)

    M.construct(
        (
            [0.1, 1.44],
            [p0, p1],
        )
    )
    M.finalize()

    # We have to do it after to ensure we correctly populate NC+Nambu
    M = M.astype(dtype)

    allclose = sisl_allclose[M.dtype]

    cFsF = {"conjugate": False, "spin": False}
    cTsF = {"conjugate": True, "spin": False}
    cFsT = {"conjugate": False, "spin": True}
    cTsT = {"conjugate": True, "spin": True}

    MT = M.transpose(**cFsF)
    MH = M.transpose(**cTsT)
    MS = M.transpose(**cFsT)
    MC = M.transpose(**cTsF)

    # just do the same op, twice, should back-transform to the same matrix
    # just do the same op, twice, should back-transform to the same matrix
    assert allclose((M - MT.transpose(**cFsF))._csr._D, 0)
    assert allclose((M - MS.transpose(**cFsT))._csr._D, 0)
    assert allclose((M - MH.transpose(**cTsT))._csr._D, 0)
    assert allclose((M - MC.transpose(**cTsF))._csr._D, 0)

    # And different conversions
    # MT ^H = M (spin-transpose and conjugate)
    assert allclose((MT.transpose(**cTsT) - MS.transpose(**cTsF))._csr._D, 0)
    assert allclose((MT.transpose(**cTsT) - MH.transpose(**cFsF))._csr._D, 0)
    assert allclose((MT.transpose(**cTsT) - MC.transpose(**cFsT))._csr._D, 0)
    # MS ^* = M (spin-transpose and conjugate)
    assert allclose((MS.transpose(**cTsF) - MT.transpose(**cTsT))._csr._D, 0)
    assert allclose((MS.transpose(**cTsF) - MH.transpose(**cFsF))._csr._D, 0)
    assert allclose((MS.transpose(**cTsF) - MC.transpose(**cFsT))._csr._D, 0)


def test_sparse_orbital_transform_ortho_unpolarized():
    M = SparseOrbitalBZSpin(geom.graphene(), spin="unpolarized")
    a = np.arange(M.spin.size(M.dtype)) + 0.3
    M.construct(([0.1, 1.44], [a, a + 0.1]))
    M.finalize()
    Mcsr = [M.tocsr(i) for i in range(M.shape[2])]

    Mt = M.transform(spin="unpolarized")
    assert np.abs(Mcsr[0] - Mt.tocsr(0)).sum() == 0

    Mt = M.transform(spin="polarized")
    assert np.abs(Mcsr[0] - Mt.tocsr(0)).sum() == 0
    assert np.abs(Mcsr[0] - Mt.tocsr(1)).sum() == 0

    Mt = M.transform(spin="non-colinear")
    assert np.abs(Mcsr[0] - Mt.tocsr(0)).sum() == 0
    assert np.abs(Mcsr[0] - Mt.tocsr(1)).sum() == 0
    assert np.abs(Mt.tocsr(2)).sum() == 0
    assert np.abs(Mt.tocsr(-1)).sum() == 0

    for s in ("so", "nambu"):
        Mt = M.transform(spin=s)
        assert np.abs(Mcsr[0] - Mt.tocsr(0)).sum() == 0
        assert np.abs(Mcsr[0] - Mt.tocsr(1)).sum() == 0
        assert np.abs(Mt.tocsr(2)).sum() == 0
        assert np.abs(Mt.tocsr(-1)).sum() == 0


def test_sparse_orbital_transform_nonortho_unpolarized():
    M = SparseOrbitalBZSpin(geom.graphene(), spin="unpolarized", orthogonal=False)
    a = np.arange(M.spin.size(M.dtype) + 1) + 0.3
    M.construct(([0.1, 1.44], [a, a + 0.1]))
    M.finalize()
    Mcsr = [M.tocsr(i) for i in range(M.shape[2])]

    Mt = M.transform(spin="unpolarized")
    assert np.abs(Mcsr[0] - Mt.tocsr(0)).sum() == 0
    assert np.abs(Mcsr[-1] - Mt.tocsr(-1)).sum() == 0

    Mt = M.transform(spin="polarized")
    assert np.abs(Mcsr[0] - Mt.tocsr(0)).sum() == 0
    assert np.abs(Mcsr[0] - Mt.tocsr(1)).sum() == 0
    assert np.abs(Mcsr[-1] - Mt.tocsr(-1)).sum() == 0

    Mt = M.transform(spin="non-colinear")
    assert np.abs(Mcsr[0] - Mt.tocsr(0)).sum() == 0
    assert np.abs(Mcsr[0] - Mt.tocsr(1)).sum() == 0
    assert np.abs(Mt.tocsr(2)).sum() == 0
    assert np.abs(Mcsr[-1] - Mt.tocsr(-1)).sum() == 0

    for s in ("so", "nambu"):
        Mt = M.transform(spin=s)
        assert np.abs(Mcsr[0] - Mt.tocsr(0)).sum() == 0
        assert np.abs(Mcsr[0] - Mt.tocsr(1)).sum() == 0
        assert np.abs(Mt.tocsr(2)).sum() == 0
        assert np.abs(Mcsr[-1] - Mt.tocsr(-1)).sum() == 0


def test_sparse_orbital_transform_ortho_polarized():
    M = SparseOrbitalBZSpin(geom.graphene(), spin="polarized")
    a = np.arange(M.spin.size(M.dtype)) + 0.3
    M.construct(([0.1, 1.44], [a, a + 0.1]))
    M.finalize()
    Mcsr = [M.tocsr(i) for i in range(M.shape[2])]

    Mt = M.transform(spin="unpolarized")
    assert np.abs(0.5 * Mcsr[0] + 0.5 * Mcsr[1] - Mt.tocsr(0)).sum() == 0

    Mt = M.transform(spin="polarized")
    assert np.abs(Mcsr[0] - Mt.tocsr(0)).sum() == 0
    assert np.abs(Mcsr[1] - Mt.tocsr(1)).sum() == 0

    Mt = M.transform(spin="non-colinear")
    assert np.abs(Mcsr[0] - Mt.tocsr(0)).sum() == 0
    assert np.abs(Mcsr[1] - Mt.tocsr(1)).sum() == 0
    assert np.abs(Mt.tocsr(2)).sum() == 0
    assert np.abs(Mt.tocsr(-1)).sum() == 0

    for s in ("so", "nambu"):
        Mt = M.transform(spin=s)
        assert np.abs(Mcsr[0] - Mt.tocsr(0)).sum() == 0
        assert np.abs(Mcsr[1] - Mt.tocsr(1)).sum() == 0
        assert np.abs(Mt.tocsr(2)).sum() == 0
        assert np.abs(Mt.tocsr(-1)).sum() == 0


def test_sparse_orbital_transform_ortho_nc():
    M = SparseOrbitalBZSpin(geom.graphene(), spin="non-colinear")
    a = np.arange(M.spin.size(M.dtype)) + 0.3
    M.construct(([0.1, 1.44], [a, a + 0.1]))
    M.finalize()
    Mcsr = [M.tocsr(i) for i in range(M.shape[2])]

    Mt = M.transform(spin="unpolarized")
    assert np.abs(0.5 * Mcsr[0] + 0.5 * Mcsr[1] - Mt.tocsr(0)).sum() == 0

    Mt = M.transform(spin="polarized")
    assert np.abs(Mcsr[0] - Mt.tocsr(0)).sum() == 0
    assert np.abs(Mcsr[1] - Mt.tocsr(1)).sum() == 0

    Mt = M.transform(spin="non-colinear")
    assert np.abs(Mcsr[0] - Mt.tocsr(0)).sum() == 0
    assert np.abs(Mcsr[1] - Mt.tocsr(1)).sum() == 0
    assert np.abs(Mcsr[2] - Mt.tocsr(2)).sum() == 0
    assert np.abs(Mcsr[3] - Mt.tocsr(3)).sum() == 0

    for s in ("so", "nambu"):
        Mt = M.transform(spin=s)
        assert np.abs(Mcsr[0] - Mt.tocsr(0)).sum() == 0
        assert np.abs(Mcsr[1] - Mt.tocsr(1)).sum() == 0
        assert np.abs(Mcsr[2] - Mt.tocsr(2)).sum() == 0
        assert np.abs(Mcsr[3] - Mt.tocsr(3)).sum() == 0


@pytest.mark.filterwarnings("ignore", message="*non-Hermitian on-site")
def test_sparse_orbital_transform_ortho_so():
    M = SparseOrbitalBZSpin(geom.graphene(), spin="so")
    a = np.arange(M.spin.size(M.dtype)) + 0.3
    M.construct(([0.1, 1.44], [a, a + 0.1]))
    M.finalize()
    Mcsr = [M.tocsr(i) for i in range(M.shape[2])]

    Mt = M.transform(spin="unpolarized")
    assert np.abs(0.5 * Mcsr[0] + 0.5 * Mcsr[1] - Mt.tocsr(0)).sum() == 0

    Mt = M.transform(spin="polarized")
    assert np.abs(Mcsr[0] - Mt.tocsr(0)).sum() == 0
    assert np.abs(Mcsr[1] - Mt.tocsr(1)).sum() == 0

    Mt = M.transform(spin="non-colinear")
    assert np.abs(Mcsr[0] - Mt.tocsr(0)).sum() == 0
    assert np.abs(Mcsr[1] - Mt.tocsr(1)).sum() == 0
    assert np.abs(Mcsr[2] - Mt.tocsr(2)).sum() == 0
    assert np.abs(Mcsr[3] - Mt.tocsr(3)).sum() == 0

    for s in ("so", "nambu"):
        Mt = M.transform(spin=s)
        assert np.abs(Mcsr[0] - Mt.tocsr(0)).sum() == 0
        assert np.abs(Mcsr[1] - Mt.tocsr(1)).sum() == 0
        assert np.abs(Mcsr[2] - Mt.tocsr(2)).sum() == 0
        assert np.abs(Mcsr[3] - Mt.tocsr(3)).sum() == 0


@pytest.mark.filterwarnings("ignore", message="*non-Hermitian on-site")
def test_sparse_orbital_transform_nonortho_so():
    M = SparseOrbitalBZSpin(geom.graphene(), spin="so", orthogonal=False)
    a = np.arange(M.spin.size(M.dtype) + 1) + 0.3
    M.construct(([0.1, 1.44], [a, a + 0.1]))
    M.finalize()
    Mcsr = [M.tocsr(i) for i in range(M.shape[2])]

    Mt = M.transform(spin="unpolarized")
    assert np.abs(0.5 * Mcsr[0] + 0.5 * Mcsr[1] - Mt.tocsr(0)).sum() == 0
    assert np.abs(Mcsr[-1] - Mt.tocsr(-1)).sum() == 0

    Mt = M.transform(spin="polarized")
    assert np.abs(Mcsr[0] - Mt.tocsr(0)).sum() == 0
    assert np.abs(Mcsr[1] - Mt.tocsr(1)).sum() == 0
    assert np.abs(Mcsr[-1] - Mt.tocsr(-1)).sum() == 0

    Mt = M.transform(spin="non-colinear")
    assert np.abs(Mcsr[0] - Mt.tocsr(0)).sum() == 0
    assert np.abs(Mcsr[1] - Mt.tocsr(1)).sum() == 0
    assert np.abs(Mcsr[2] - Mt.tocsr(2)).sum() == 0
    assert np.abs(Mcsr[3] - Mt.tocsr(3)).sum() == 0
    assert np.abs(Mcsr[-1] - Mt.tocsr(-1)).sum() == 0

    for s in ("so", "nambu"):
        Mt = M.transform(spin=s)
        assert np.abs(Mcsr[0] - Mt.tocsr(0)).sum() == 0
        assert np.abs(Mcsr[1] - Mt.tocsr(1)).sum() == 0
        assert np.abs(Mcsr[2] - Mt.tocsr(2)).sum() == 0
        assert np.abs(Mcsr[3] - Mt.tocsr(3)).sum() == 0
        assert np.abs(Mcsr[-1] - Mt.tocsr(-1)).sum() == 0


def test_sparse_orbital_transform_basis():
    M = SparseOrbitalBZSpin(geom.graphene(), spin="polarized", orthogonal=False)
    M.construct(([0.1, 1.44], [(3.0, 2.0, 1.0), (0.3, 0.2, 0.0)]))
    M.finalize()
    Mcsr = [M.tocsr(i) for i in range(M.shape[2])]

    Mt = M.transform(orthogonal=True).transform(orthogonal=False)
    assert M.dim == Mt.dim
    assert np.abs(Mcsr[0] - Mt.tocsr(0)).sum() == 0
    assert np.abs(Mcsr[1] - Mt.tocsr(1)).sum() == 0
    assert np.abs(Mcsr[-1] - Mt.tocsr(-1)).sum() == 0


@pytest.mark.skipif(
    sys.platform.startswith("win") or "arm" in platform.machine().lower(),
    reason="Data type cannot be float128",
)
def test_sparse_orbital_transform_combinations():
    M = SparseOrbitalBZSpin(
        geom.graphene(), spin="polarized", orthogonal=False, dtype=np.int32
    )
    M.construct(([0.1, 1.44], [(3, 2, 1), (2, 1, 0)]))
    M.finalize()
    Mcsr = [M.tocsr(i) for i in range(M.shape[2])]

    Mt = (
        M.transform(spin="non-colinear", orthogonal=True)
        .astype(np.float64)
        .transform(spin="polarized", orthogonal=False)
    )
    assert M.dim == Mt.dim
    assert np.abs(Mcsr[0] - Mt.tocsr(0)).sum() == 0
    assert np.abs(Mcsr[1] - Mt.tocsr(1)).sum() == 0
    assert np.abs(Mcsr[-1] - Mt.tocsr(-1)).sum() == 0

    Mt = (
        M.transform(orthogonal=True)
        .astype(np.float128)
        .transform(spin="so", orthogonal=False)
        .astype(np.float64)
    )
    assert np.abs(Mcsr[0] - Mt.tocsr(0)).sum() == 0
    assert np.abs(Mcsr[1] - Mt.tocsr(1)).sum() == 0
    assert np.abs(Mt.tocsr(2)).sum() == 0
    assert np.abs(Mcsr[-1] - Mt.tocsr(-1)).sum() == 0

    Mt = (
        M.transform(spin="polarized", orthogonal=True)
        .transform(spin="so", orthogonal=False)
        .astype(np.float64)
    )
    assert np.abs(Mcsr[0] - Mt.tocsr(0)).sum() == 0
    assert np.abs(Mcsr[1] - Mt.tocsr(1)).sum() == 0
    assert np.abs(Mt.tocsr(2)).sum() == 0
    assert np.abs(Mcsr[-1] - Mt.tocsr(-1)).sum() == 0

    Mt = (
        M.transform(spin="unpolarized", orthogonal=True)
        .astype(np.float32)
        .transform(orthogonal=False)
        .astype(np.complex128)
    )
    assert np.abs(0.5 * Mcsr[0] + 0.5 * Mcsr[1] - Mt.tocsr(0)).sum() == 0


def test_sparse_orbital_transform_matrix():
    M = SparseOrbitalBZSpin(
        geom.graphene(), spin="polarized", orthogonal=False, dtype=np.int32
    )
    M.construct(([0.1, 1.44], [(1, 2, 3), (4, 5, 6)]))
    M.finalize()
    Mcsr = [M.tocsr(i) for i in range(M.shape[2])]

    Mt = M.transform(spin="unpolarized", matrix=np.ones((1, 3)), orthogonal=True)
    assert Mt.dim == 1
    assert np.abs(Mcsr[0] + Mcsr[1] + Mcsr[2] - Mt.tocsr(0)).sum() == 0

    Mt = M.transform(spin="polarized", matrix=np.ones((2, 3)), orthogonal=True)
    assert Mt.dim == 2
    assert np.abs(Mcsr[0] + Mcsr[1] + Mcsr[2] - Mt.tocsr(1)).sum() == 0

    Mt = M.transform(matrix=np.ones((3, 3)))
    assert Mt.dim == 3
    assert np.abs(Mcsr[0] + Mcsr[1] + Mcsr[2] - Mt.tocsr(2)).sum() == 0

    Mt = M.transform(spin="non-colinear", matrix=np.ones((4, 3)), orthogonal=True)
    assert Mt.dim == 4
    assert np.abs(Mcsr[0] + Mcsr[1] + Mcsr[2] - Mt.tocsr(3)).sum() == 0

    Mt = M.transform(spin="non-colinear", matrix=np.ones((5, 3)))
    assert Mt.dim == 5
    assert np.abs(Mcsr[0] + Mcsr[1] + Mcsr[2] - Mt.tocsr(4)).sum() == 0

    Mt = M.transform(spin="so", matrix=np.ones((8, 3)), orthogonal=True)
    assert Mt.dim == 8
    assert np.abs(Mcsr[0] + Mcsr[1] + Mcsr[2] - Mt.tocsr(7)).sum() == 0

    Mt = M.transform(spin="so", matrix=np.ones((9, 3)))
    assert Mt.dim == 9
    assert np.abs(Mcsr[0] + Mcsr[1] + Mcsr[2] - Mt.tocsr(8)).sum() == 0


def test_sparse_orbital_transform_fail():
    M = SparseOrbitalBZSpin(
        geom.graphene(), spin="polarized", orthogonal=False, dtype=np.int32
    )
    M.construct(([0.1, 1.44], [(1, 2, 3), (4, 5, 6)]))
    M.finalize()

    with pytest.raises(ValueError):
        M.transform(np.zeros([2, 2]), spin="unpolarized")


@pytest.mark.parametrize("dtype", [np.float32, np.float64, np.complex64, np.complex128])
@pytest.mark.parametrize(
    "spin", ["unpolarized", "polarized", "non-colinear", "spin-orbit"]
)
def test_sparseorbital_spin_dtypes(dtype, spin):
    gr = geom.graphene()

    M = SparseOrbitalBZSpin(gr, spin=Spin(spin), dtype=dtype)
    assert M.dtype == dtype


def test_sparseorbital_fromsp_csr_matrix():
    gr = geom.graphene()
    no, no_s = gr.no, gr.no_s
    s1 = sps.csr_matrix((no, no_s))

    M = SparseOrbitalBZ.fromsp(gr, s1)
    assert M.shape == (no, no_s, 1)
    assert M.nnz == 0

    M = SparseOrbitalBZ.fromsp(gr, [s1, s1], orthogonal=False)
    assert M.shape == (no, no_s, 2)

    M = SparseOrbitalBZ.fromsp(gr, [s1, s1], S=s1, orthogonal=False)
    assert M.shape == (no, no_s, 3)


def test_sparseorbital_fromsp_sparsecsr():
    gr = geom.graphene()
    no, no_s = gr.no, gr.no_s
    s1 = SparseCSR((no, no_s, 2))

    M = SparseOrbitalBZ.fromsp(gr, s1)
    assert M.shape == (no, no_s, 2)
    assert M.nnz == 0

    M = SparseOrbitalBZ.fromsp(gr, [s1, s1])
    assert M.shape == (no, no_s, 4)

    M = SparseOrbitalBZ.fromsp(gr, [s1, s1], orthogonal=False)
    assert M.shape == (no, no_s, 4)


def test_sparseorbital_fromsp_sparsecsr_overlap_error():
    gr = geom.graphene()
    no, no_s = gr.no, gr.no_s
    s1 = SparseCSR((no, no_s, 2))
    with pytest.raises(ValueError):
        SparseOrbitalBZ.fromsp(gr, s1, S=s1)


def test_sparseorbital_fromsp_combined():
    gr = geom.graphene()
    no, no_s = gr.no, gr.no_s
    s1 = SparseCSR((no, no_s, 2))
    s2 = sps.csr_matrix((no, no_s))

    M = SparseOrbitalBZ.fromsp(gr, [s1, s2])
    assert M.shape == (no, no_s, 3)
    assert M.nnz == 0

    M = SparseOrbitalBZ.fromsp(gr, s1, S=s2)
    assert M.shape == (no, no_s, 3)

    M = SparseOrbitalBZ.fromsp(gr, [s1, s2], S=s2)
    assert M.shape == (no, no_s, 4)


def test_sparseorbital_fromsp_orthogonal():
    gr = geom.graphene()
    no, no_s = gr.no, gr.no_s
    s1 = SparseCSR((no, no_s, 2))
    s2 = sps.csr_matrix((no, no_s))

    M = SparseOrbitalBZ.fromsp(gr, [s1, s2], orthogonal=False)
    assert M.shape == (no, no_s, 3)
    assert M.nnz == 0
    assert not M.orthogonal

    M1 = SparseOrbitalBZ.fromsp(gr, M)
    assert M1.shape == (no, no_s, 2)
    assert M1.orthogonal

    # we only extract the matrix elements (not overlap)
    # and then add the overlap part explicitly
    M1 = SparseOrbitalBZ.fromsp(gr, M, S=M)
    assert M1.shape == (no, no_s, 3)
    assert not M1.orthogonal



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import pytest

from sisl import Spin
from sisl.viz.data import PDOSData
from sisl.viz.plots import pdos_plot

pytestmark = [pytest.mark.viz, pytest.mark.plots]


@pytest.fixture(scope="module", params=["plotly", "matplotlib"])
def backend(request):
    pytest.importorskip(request.param)
    return request.param


@pytest.fixture(
    scope="module", params=["unpolarized", "polarized", "noncolinear", "spinorbit"]
)
def spin(request):
    return Spin(request.param)


@pytest.fixture(scope="module")
def pdos_data(spin):
    return PDOSData.toy_example(spin=spin)


def test_pdos_plot(pdos_data, backend):
    pdos_plot(pdos_data, backend=backend)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

from sisl.shape import *

pytestmark = pytest.mark.shape


def test_binary_op():
    e = Ellipsoid(1.0)
    s = Sphere(1.0)

    new = e + s
    str(new)

    new = new - new
    new = new & (new | e) ^ s
    new.center
    new2 = new.translate([0, 1, 2])
    assert not np.allclose(new.center, new2.center)

    assert new.volume < 0.0
    str(new)


def test_null():
    null = NullShape()
    assert null.volume == 0.0
    assert len(null.within_index(np.random.rand(1000, 3))) == 0

    assert null.to.Ellipsoid().volume < 1e-64
    assert null.to.Cuboid().volume < 1e-64
    assert null.to.Sphere().volume < 1e-64


def test_binary_op_within():
    e = Ellipsoid(0.5)
    c = Cube(1.0)

    xc = [0.499] * 3
    new = e + c
    assert not e.within(xc)
    assert c.within(xc)
    assert new.within(xc)

    xe = [0.0] * 3
    new = e - c
    assert e.within(xe)
    assert c.within(xe)
    assert not new.within(xe)

    new = e & c
    assert not e.within(xc)
    assert c.within(xc)
    assert not new.within(xc)
    assert e.within(xe)
    assert c.within(xe)
    assert new.within(xe)

    # e ^ c == c ^ e
    new = e ^ c
    assert not e.within(xc)
    assert c.within(xc)
    assert new.within(xc)
    assert e.within(xe)
    assert c.within(xe)
    assert not new.within(xe)

    new = c ^ e
    assert new.within(xc)
    assert not new.within(xe)


def test_binary_op_toSphere():
    e = Ellipsoid(0.5)
    c = Cube(1.0)

    r = 0.5 * 3**0.5
    new = e + c
    assert new.to.Sphere().radius.max() == pytest.approx(r)

    new = e - c
    assert new.to.Sphere().radius.max() == pytest.approx(r)

    # with the AND operator we can reduce to smallest enclosed sphere
    new = e & c
    assert new.to.Sphere().radius.max() == pytest.approx(0.5)

    # e ^ c == c ^ e
    new = e ^ c
    assert new.to.Sphere().radius.max() == pytest.approx(r)

    new = c ^ e
    assert new.to.Sphere().radius.max() == pytest.approx(r)

    new = c ^ e
    assert new.scale(2).to.Sphere().radius.max() == pytest.approx(r * 2)


def test_toSphere_and():
    left = Sphere(1.0)
    right = Sphere(1.0, center=[0.6] * 3)

    new = left & right
    s = new.to.Sphere()
    assert s.radius.max() < 0.9

    left = Sphere(2.0)
    right = Sphere(1.0, center=[0.5] * 3)

    new = left & right
    s = new.to.Sphere()
    assert s.radius.max() == pytest.approx(1.0)

    left = Sphere(2.0, center=[10, 10, 10])
    right = Sphere(1.0, center=[10.5] * 3)

    new = left & right
    s2 = new.to.Sphere()
    assert s2.radius.max() == pytest.approx(1.0)
    # Assert it also works for displaced centers
    assert np.allclose(s.radius, s2.radius)
    assert np.allclose(s.center, s2.center - 10)

    left = Sphere(2.0)
    right = Sphere(1.0, center=[10.5] * 3)

    new = left & right
    s = new.to.Sphere()
    assert s.radius.max() < 0.01


def test_toEllipsoid_and():
    left = Ellipsoid(1.0)
    right = Ellipsoid(1.0, center=[0.6] * 3)

    new = left & right
    s = new.to.Ellipsoid()
    assert s.radius.max() < 0.9

    left = Ellipsoid(2.0)
    right = Ellipsoid(1.0, center=[0.5] * 3)

    new = left & right
    s = new.to.Ellipsoid()
    assert s.radius.max() == pytest.approx(1.0)

    left = Ellipsoid(2.0, center=[10, 10, 10])
    right = Ellipsoid(1.0, center=[10.5] * 3)

    new = left & right
    s2 = new.to.Ellipsoid()
    assert s2.radius.max() == pytest.approx(1.0)
    # Assert it also works for displaced centers
    assert np.allclose(s.radius, s2.radius)
    assert np.allclose(s.center, s2.center - 10)

    left = Ellipsoid(2.0)
    right = Ellipsoid(1.0, center=[10.5] * 3)

    new = left & right
    s = new.to.Ellipsoid()
    assert s.radius.max() < 0.01


def test_toCuboid_and():
    left = Cuboid(1.0)
    right = Cuboid(1.0, center=[0.6] * 3)

    new = left & right
    s = new.to.Cuboid()
    assert s.edge_length.max() < 0.9 * 2

    left = Cuboid(2.0)
    right = Cuboid(1.0, center=[0.5] * 3)

    new = left & right
    s = new.to.Cuboid()
    assert s.edge_length.max() >= 1.5

    left = Cuboid(2.0, center=[10, 10, 10])
    right = Cuboid(1.0, center=[10.5] * 3)

    new = left & right
    s2 = new.to.Cuboid()
    assert s2.edge_length.max() > 1.5
    # Assert it also works for displaced centers
    assert np.allclose(s.edge_length, s2.edge_length)
    assert np.allclose(s.center, s2.center - 10)

    left = Cuboid(2.0)
    right = Cuboid(1.0, center=[10.5] * 3)

    new = left & right
    s = new.to.Cuboid()
    assert s.edge_length.max() < 0.01



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import pytest

import sisl
from sisl.viz.data import PDOSData

pytestmark = [pytest.mark.viz, pytest.mark.data]


@pytest.mark.parametrize(
    "spin", ["unpolarized", "polarized", "noncolinear", "spinorbit"]
)
def test_pdos_from_sisl_H(spin):
    gr = sisl.geom.graphene()
    H = sisl.Hamiltonian(gr)
    H.construct([(0.1, 1.44), (0, -2.7)])

    n_spin, H = {
        "unpolarized": (1, H),
        "polarized": (2, H.transform(spin=sisl.Spin.POLARIZED)),
        "noncolinear": (4, H.transform(spin=sisl.Spin.NONCOLINEAR)),
        "spinorbit": (4, H.transform(spin=sisl.Spin.SPINORBIT)),
    }[spin]

    data = PDOSData.new(H, Erange=(-5, 5))

    checksum = 17.599343960516066
    if n_spin > 1:
        checksum = checksum * 2

    data.sanity_check(
        na=2, no=2, n_spin=n_spin, atom_tags=("C",), dos_checksum=checksum
    )


@pytest.mark.parametrize("spin", ["unpolarized", "polarized", "non-collinear"])
def test_pdos_from_siesta_PDOS(spin, sisl_files):
    n_spin, filename = {
        "unpolarized": (1, "SrTiO3.PDOS"),
        "polarized": (2, "SrTiO3.PDOS"),
        "non-collinear": (4, "SrTiO3.PDOS"),
    }[spin]

    file = sisl_files("siesta", "SrTiO3", spin, filename)

    data = PDOSData.new(file)

    checksum = 2376.8803000000003 / 2
    if n_spin > 1:
        checksum = checksum * 2

    data.sanity_check(
        na=5, no=72, n_spin=n_spin, atom_tags=("Sr", "Ti", "O"), dos_checksum=checksum
    )


def test_pdos_from_siesta_wfsx(sisl_files):
    nspin = 4
    dir = "Bi2Se3_3layer"

    # From a siesta .WFSX file
    # Since there is no hamiltonian for bi2se3_3ql.fdf, we create a dummy one
    wfsx = sisl.get_sile(sisl_files("siesta", dir, "Bi2Se3.bands.WFSX"))

    geometry = sisl.get_sile(sisl_files("siesta", dir, "Bi2Se3.fdf")).read_geometry()
    geometry = sisl.Geometry(geometry.xyz, atoms=wfsx.read_basis())

    H = sisl.Hamiltonian(geometry, dim=nspin)

    data = PDOSData.new(wfsx, H=H)

    # For now, the checksum is 0 because we have no overlap matrix.
    checksum = 0

    data.sanity_check(
        na=15, no=195, n_spin=nspin, atom_tags=("Bi", "Se"), dos_checksum=checksum
    )


@pytest.mark.parametrize(
    "spin", ["unpolarized", "polarized", "noncolinear", "spinorbit"]
)
def test_toy_example(spin):
    data = PDOSData.toy_example(spin=spin)

    n_spin = {"unpolarized": 1, "polarized": 2, "noncolinear": 4, "spinorbit": 4}[spin]

    data.sanity_check(
        n_spin=n_spin, na=data.geometry.na, no=data.geometry.no, atom_tags=["C"]
    )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

from sisl._namedindex import NamedIndex

pytestmark = pytest.mark.namedindex


def test_ni_init():
    ni = NamedIndex()
    str(ni)
    ni = NamedIndex("name", [1])
    str(ni)
    ni = NamedIndex(["name-1", "name-2"], [[1], [0]])


def test_ni_iter():
    ni = NamedIndex()
    assert len(ni) == 0
    ni.add_name("name-1", [0])
    assert len(ni) == 1
    ni.add_name("name-2", [1])
    assert len(ni) == 2
    for n in ni:
        assert n in ["name-1", "name-2"]
    assert "name-1" in ni
    assert "name-2" in ni


def test_ni_clear():
    ni = NamedIndex()
    assert len(ni) == 0
    ni.add_name("name-1", [0])
    assert len(ni) == 1
    ni.clear()
    assert len(ni) == 0


def test_ni_copy():
    ni = NamedIndex()
    ni.add_name("name-1", [0])
    ni.add_name("name-2", [1])
    n2 = ni.copy()
    assert ni._name == n2._name


def test_ni_delete():
    ni = NamedIndex()
    ni.add_name("name-1", [0])
    ni.add_name("name-2", [1])
    ni.delete_name("name-1")
    for n in ni:
        assert n in ["name-2"]


def test_ni_items():
    ni = NamedIndex()
    ni["Hello"] = [0]
    ni[[1, 2]] = "Hello-1"
    assert np.all(ni["Hello"] == [0])
    no = ni.remove_index(1)
    assert np.all(no["Hello-1"] == [2])


def test_ni_dict():
    ni = NamedIndex({"r1": [1, 2], "r2": [3, 4]})
    assert len(ni) == 2
    assert np.all(ni["r1"] == [1, 2])
    assert np.all(ni["r2"] == [3, 4])


def test_ni_sub_index():
    ni = NamedIndex({"r1": [1, 2], "r2": [3, 4]})
    no = ni.sub_index([2, 4])
    assert len(no) == 2
    assert len(no["r1"]) == 1
    assert len(no["r2"]) == 1

    no = ni.sub_name("r1")
    assert no.names[0] == "r1"
    assert len(no) == 1


def test_ni_merge():
    ni1 = NamedIndex(["r1", "r2"], [[1, 2], [3, 4]])
    ni2 = ni1.copy()
    ni2.add_name("r3", [5, 6])

    with pytest.raises(ValueError):
        ni1.merge(ni2)
    with pytest.raises(ValueError):
        ni1.merge(ni2, duplicate="raise")
    with pytest.raises(ValueError):
        ni1.merge(ni2, duplicate="not something viable")

    ni3 = ni1.merge(ni2, offset=10, duplicate="union")
    assert len(ni3) == 3
    assert len(ni3["r1"]) == 4
    assert len(ni3["r2"]) == 4
    assert len(ni3["r3"]) == 2

    ni3 = ni1.merge(ni2, offset=10, duplicate="omit")
    assert len(ni3) == 1

    ni3 = ni1.merge(ni2, offset=10, duplicate="left")
    assert len(ni3) == 3
    assert np.array_equal(ni3["r1"], ni1["r1"])
    assert np.array_equal(ni3["r2"], ni1["r2"])
    assert np.array_equal(ni3["r3"], ni2["r3"] + 10)

    ni3 = ni1.merge(ni2, offset=10, duplicate="right")
    assert len(ni3) == 3
    assert np.array_equal(ni3["r1"], ni2["r1"] + 10)
    assert np.array_equal(ni3["r2"], ni2["r2"] + 10)
    assert np.array_equal(ni3["r3"], ni2["r3"] + 10)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import sys
import tempfile
import zipfile

import numpy as np
import pytest

import sisl
from sisl.io._zipfile import ZipPath

# We will xfail all tests if the python version is <=3.9,
# this is because sisl is expected to raise a RuntimeError
# explaining that zipfile functionality requires Python 3.10 or newer.
_is_old_python = sys.version_info < (3, 10)
# 3.9 also prints very ugly errors when deleting a zipfile which make
# it difficult to read the test output. So we patch the __del__ method.
if _is_old_python:
    zipfile.ZipFile.__del__ = lambda self: None


@pytest.mark.xfail(
    _is_old_python,
    reason="Zip file functionality requires Python 3.10 or newer",
    raises=RuntimeError,
)
def test_zipfile_preserved():
    """Test that the zipfile is preserved through the sile framework

    This is VERY important, because the lookup for paths will only be
    fast if the index is already built. If each time a new sile is
    created we need to create a new zipfile, it will be very slow.
    """
    # Write a temporary zipfile
    _, tempfile_path = tempfile.mkstemp(suffix=".zip")
    with zipfile.ZipFile(tempfile_path, "a") as f:
        ...

    f = zipfile.ZipFile(tempfile_path)

    root_path = zipfile.Path(f, "")

    fdf = sisl.get_sile(root_path / "test.fdf")

    assert isinstance(fdf.file, ZipPath)
    assert fdf.file.root is f


@pytest.mark.xfail(
    _is_old_python,
    reason="Zip file functionality requires Python 3.10 or newer",
    raises=RuntimeError,
)
@pytest.mark.parametrize("specify_class", [True, False])
@pytest.mark.parametrize("external_zip", [True, False])
def test_zipfile_write_read(external_zip: bool, specify_class: bool):
    """Test that we can write and read within a zipfile.

    If the user provides the zipfile path, instead of
    letting the sile framework build it, the sile should not
    close the zipfile.

    The user is then responsible for closing the zipfile. This
    is important because if the sile closes the zipfile you can't
    write to it anymore.
    """

    geometry = sisl.geom.graphene()

    # Write a temporary zipfile
    _, tempfile_path = tempfile.mkstemp(suffix=".zip")
    zip_file = zipfile.ZipFile(tempfile_path, "a")

    filename = "graphene.dat{xyzSile}" if specify_class else "graphene.xyz"

    if external_zip:
        graphene_xyz = zipfile.Path(zip_file, filename)
    else:
        graphene_xyz = tempfile_path + "/" + filename

    geometry.write(graphene_xyz)

    if external_zip:
        # Try to write another file
        zipfile.Path(zip_file, "other_file.txt").open("w").write("Hello world!")

        # Close the zipfile
        zip_file.close()

    # Make sisl read from the written zipfile
    graphene_xyz = str(graphene_xyz)
    read_geometry = sisl.get_sile(graphene_xyz).read_geometry()

    assert np.allclose(read_geometry.xyz, geometry.xyz)
    assert np.allclose(read_geometry.cell, geometry.cell)


@pytest.mark.xfail(
    _is_old_python,
    reason="Zip file functionality requires Python 3.10 or newer",
    raises=RuntimeError,
)
@pytest.mark.parametrize("external_zip", [True, False])
@pytest.mark.parametrize("from_fdf", [True, False])
def test_zipfile_write_read_binary(external_zip: bool, from_fdf: bool):
    """Test that we can write and read binary files within a zipfile"""

    geometry = sisl.geom.graphene()
    H = sisl.Hamiltonian(geometry)
    H[0, 0] = 1.3
    H[1, 1] = 1.5

    # Write a temporary zipfile
    _, tempfile_path = tempfile.mkstemp(suffix=".zip")
    zip_file = zipfile.ZipFile(tempfile_path, "a")

    H_filename = "siesta.HSX"

    if external_zip:
        H_path = zipfile.Path(zip_file, H_filename)
        fdf_path = zipfile.Path(zip_file, "test.fdf")
    else:
        H_path = tempfile_path + "/" + H_filename
        fdf_path = tempfile_path + "/test.fdf"

    if from_fdf:
        # Write the geometry to the fdf file
        geometry.write(fdf_path)

    H.write(H_path)

    if external_zip:
        # Try to write another file
        zipfile.Path(zip_file, "other_file.txt").open("w").write("Hello world!")

        # Close the zipfile
        zip_file.close()

    # Make sisl read from the written zipfile
    path = str(fdf_path) if from_fdf else str(H_path)
    read_H = sisl.get_sile(path).read_hamiltonian()

    assert np.allclose(H.tocsr().toarray(), read_H.tocsr().toarray())


@pytest.mark.xfail(
    _is_old_python,
    reason="Zip file functionality requires Python 3.10 or newer",
    raises=RuntimeError,
)
@pytest.mark.parametrize("external_zip", [True, False])
@pytest.mark.parametrize("from_fdf", [True, False])
def test_zipfile_write_read_cdf(external_zip: bool, from_fdf: bool):
    """Test that we can write and read CDF files within a zipfile"""
    pytest.importorskip("netCDF4")

    geometry = sisl.geom.graphene()
    grid = sisl.Grid((2, 2, 2), geometry=geometry)
    grid[:] = np.random.random(grid.shape)

    # Write a temporary zipfile
    _, tempfile_path = tempfile.mkstemp(suffix=".zip")
    zip_file = zipfile.ZipFile(tempfile_path, "a")

    grid_filename = "Rho.grid.nc"

    if external_zip:
        grid_path = zipfile.Path(zip_file, grid_filename)
        fdf_path = zipfile.Path(zip_file, "test.fdf")
    else:
        grid_path = tempfile_path + "/" + grid_filename
        fdf_path = tempfile_path + "/test.fdf"

    if from_fdf:
        # Write the geometry to the fdf file
        geometry.write(fdf_path)

    grid.write(grid_path)

    if external_zip:
        # Try to write another file
        zipfile.Path(zip_file, "other_file.txt").open("w").write("Hello world!")

        # Close the zipfile
        zip_file.close()

    # Make sisl read from the written zipfile
    if from_fdf:
        read_grid = sisl.get_sile(str(fdf_path)).read_grid("RHO")
    else:
        # Read the grid from the grid.nc file
        read_grid = sisl.get_sile(str(grid_path)).read_grid()

    assert np.allclose(grid.grid, read_grid.grid)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest
import xarray as xr

import sisl
from sisl.viz.processors.atom import reduce_atom_data

pytestmark = [pytest.mark.viz, pytest.mark.processors]


@pytest.fixture(scope="module")
def atom_x():
    geom = sisl.geom.graphene().tile(20, 1).tile(20, 0)

    return xr.DataArray(
        geom.xyz[:, 0], coords=[("atom", range(geom.na))], attrs={"geometry": geom}
    )


@pytest.fixture(scope="module")
def atom_xyz():
    geom = sisl.geom.graphene().tile(20, 1).tile(20, 0)

    return xr.Dataset(
        {
            "x": xr.DataArray(geom.xyz[:, 0], coords=[("atom", range(geom.na))]),
            "y": xr.DataArray(geom.xyz[:, 1], coords=[("atom", range(geom.na))]),
            "z": xr.DataArray(geom.xyz[:, 2], coords=[("atom", range(geom.na))]),
        },
        attrs={"geometry": geom},
    )


def test_reduce_atom_dataarray(atom_x):
    grouped = reduce_atom_data(
        atom_x,
        [{"atoms": [0, 1], "name": "first"}, {"atoms": [5, 6], "name": "second"}],
        reduce_func=np.sum,
        groups_dim="group",
    )

    assert isinstance(grouped, xr.DataArray)
    assert float(grouped.sel(group="first")) == np.sum(atom_x.values[0:2].sum())
    assert float(grouped.sel(group="second")) == np.sum(atom_x.values[5:7].sum())


def test_reduce_atom_dataarray_cat(atom_x):
    """We test that the atoms field is correctly sanitized using the geometry attached to the xarray object."""
    grouped = reduce_atom_data(
        atom_x,
        [
            {"atoms": {"x": (0, 10)}, "name": "first"},
            {"atoms": {"x": (10, None)}, "name": "second"},
        ],
        reduce_func=np.max,
        groups_dim="group",
    )

    assert isinstance(grouped, xr.DataArray)
    assert float(grouped.sel(group="first")) <= 10
    assert float(grouped.sel(group="second")) == atom_x.max()


def test_reduce_atom_cat_nogeom(atom_x):
    """We test that the atoms field is correctly sanitized using the geometry attached to the xarray object."""
    atom_x = atom_x.copy()
    geometry = atom_x.attrs["geometry"]

    # Remove the geometry
    atom_x.attrs = {}

    # Without a geometry, it should fail to sanitize atoms specifications
    with pytest.raises(Exception):
        grouped = reduce_atom_data(
            atom_x,
            [
                {"atoms": {"x": (0, 10)}, "name": "first"},
                {"atoms": {"x": (10, None)}, "name": "second"},
            ],
            reduce_func=np.max,
            groups_dim="group",
        )

    # If we explicitly pass the geometry it should again be able to sanitize the atoms
    grouped = reduce_atom_data(
        atom_x,
        [
            {"atoms": {"x": (0, 10)}, "name": "first"},
            {"atoms": {"x": (10, None)}, "name": "second"},
        ],
        geometry=geometry,
        reduce_func=np.max,
        groups_dim="group",
    )

    assert isinstance(grouped, xr.DataArray)
    assert float(grouped.sel(group="first")) <= 10
    assert float(grouped.sel(group="second")) == atom_x.max()


def test_reduce_atom_dataset_cat(atom_xyz):
    """We test that the atoms field is correctly sanitized using the geometry attached to the xarray object."""
    grouped = reduce_atom_data(
        atom_xyz,
        [
            {"atoms": {"x": (0, 10), "y": (1, 3)}, "name": "first"},
            {"atoms": {"x": (10, None)}, "name": "second"},
        ],
        reduce_func=np.max,
        groups_dim="group",
    )

    assert isinstance(grouped, xr.Dataset)
    assert float(grouped.sel(group="first").x) <= 10
    assert float(grouped.sel(group="second").x) == atom_xyz.x.max()
    assert float(grouped.sel(group="first").y) <= 3
    assert float(grouped.sel(group="second").y) == atom_xyz.y.max()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

from sisl import Atom, Geometry, Lattice
from sisl._core.geometry import sgeom


@pytest.fixture
def setup():
    class t:
        def __init__(self):
            bond = 1.42
            sq3h = 3.0**0.5 * 0.5
            self.lattice = Lattice(
                np.array(
                    [[1.5, sq3h, 0.0], [1.5, -sq3h, 0.0], [0.0, 0.0, 10.0]], np.float64
                )
                * bond,
                nsc=[3, 3, 1],
            )
            C = Atom(Z=6, R=[bond * 1.01] * 2)
            self.g = Geometry(
                np.array([[0.0, 0.0, 0.0], [1.0, 0.0, 0.0]], np.float64) * bond,
                atoms=C,
                lattice=self.lattice,
            )

            self.mol = Geometry([[i, 0, 0] for i in range(10)], lattice=[50])

            def sg_g(**kwargs):
                kwargs["ret_geometry"] = True
                if "geometry" not in kwargs:
                    kwargs["geometry"] = self.g.copy()
                return sgeom(**kwargs)

            self.sg_g = sg_g

            def sg_mol(**kwargs):
                kwargs["ret_geometry"] = True
                if "geometry" not in kwargs:
                    kwargs["geometry"] = self.mol.copy()
                return sgeom(**kwargs)

            self.sg_mol = sg_mol

    return t()


@pytest.mark.geometry
class TestGeometry:
    def test_help(self):
        with pytest.raises(SystemExit):
            sgeom(argv=["--help"])

    def test_version(self):
        sgeom(argv=["--version"])

    def test_cite(self):
        sgeom(argv=["--cite"])

    def test_tile1(self, setup):
        cell = np.copy(setup.g.lattice.cell)
        cell[0, :] *= 2
        for tile in ["tile 2 x", "tile-x 2"]:
            tx = setup.sg_g(argv=("--" + tile).split())
            assert np.allclose(cell, tx.lattice.cell)
        cell[1, :] *= 2
        for tile in ["tile 2 y", "tile-y 2"]:
            ty = setup.sg_g(geometry=tx, argv=("--" + tile).split())
            assert np.allclose(cell, ty.lattice.cell)
        cell[2, :] *= 2
        for tile in ["tile 2 z", "tile-z 2"]:
            tz = setup.sg_g(geometry=ty, argv=("--" + tile).split())
            assert np.allclose(cell, tz.lattice.cell)

    def test_tile2(self, setup):
        cell = np.copy(setup.g.lattice.cell)
        cell[:, :] *= 2
        for xt in ["tile 2 x", "tile-x 2"]:
            xt = "--" + xt
            for yt in ["tile 2 y", "tile-y 2"]:
                yt = "--" + yt
                for zt in ["tile 2 z", "tile-z 2"]:
                    zt = "--" + zt
                    argv = " ".join([xt, yt, zt]).split()
                    t = setup.sg_g(argv=argv)
                    assert np.allclose(cell, t.lattice.cell)

    def test_repeat1(self, setup):
        cell = np.copy(setup.g.lattice.cell)
        cell[0, :] *= 2
        for repeat in ["repeat 2 x", "repeat-x 2"]:
            tx = setup.sg_g(argv=("--" + repeat).split())
            assert np.allclose(cell, tx.lattice.cell)
        cell[1, :] *= 2
        for repeat in ["repeat 2 y", "repeat-y 2"]:
            ty = setup.sg_g(geometry=tx, argv=("--" + repeat).split())
            assert np.allclose(cell, ty.lattice.cell)
        cell[2, :] *= 2
        for repeat in ["repeat 2 z", "repeat-z 2"]:
            tz = setup.sg_g(geometry=ty, argv=("--" + repeat).split())
            assert np.allclose(cell, tz.lattice.cell)

    def test_repeat2(self, setup):
        cell = np.copy(setup.g.lattice.cell)
        cell[:, :] *= 2
        for xt in ["repeat 2 x", "repeat-x 2"]:
            xt = "--" + xt
            for yt in ["repeat 2 y", "repeat-y 2"]:
                yt = "--" + yt
                for zt in ["repeat 2 z", "repeat-z 2"]:
                    zt = "--" + zt
                    argv = " ".join([xt, yt, zt]).split()
                    t = setup.sg_g(argv=argv)
                    assert np.allclose(cell, t.lattice.cell)

    def test_sub(self, setup):
        for a, l in [("0", 1), ("0,1", 2), ("0-1", 2)]:
            g = setup.sg_g(argv=["--sub", a])
            assert len(g) == l

    def test_remove(self, setup):
        geom = setup.g.tile(2, 0).tile(2, 1)
        N = len(geom)
        for a, l in [("0", 1), ("0,1", 2), ("0-1", 2)]:
            g = setup.sg_g(geometry=geom.copy(), argv=["--remove", a])
            assert len(g) == N - l

    def test_rotation1(self, setup):
        rot = setup.sg_g(argv="--rotate 180 z".split())
        rot.lattice.cell[2, 2] *= -1
        assert np.allclose(-rot.lattice.cell, setup.g.lattice.cell)
        assert np.allclose(-rot.xyz, setup.g.xyz)

        rot = setup.sg_g(argv="--rotate-z 180".split())
        rot.lattice.cell[2, 2] *= -1
        assert np.allclose(-rot.lattice.cell, setup.g.lattice.cell)
        assert np.allclose(-rot.xyz, setup.g.xyz)

        rot = setup.sg_g(argv="--rotate rpi z".split())
        rot.lattice.cell[2, 2] *= -1
        assert np.allclose(-rot.lattice.cell, setup.g.lattice.cell)
        assert np.allclose(-rot.xyz, setup.g.xyz)

        rot = setup.sg_g(argv="--rotate-z rpi".split())
        rot.lattice.cell[2, 2] *= -1
        assert np.allclose(-rot.lattice.cell, setup.g.lattice.cell)
        assert np.allclose(-rot.xyz, setup.g.xyz)

    def test_swap(self, setup):
        s = setup.sg_g(argv="--swap 0 1".split())
        for i in [0, 1, 2]:
            assert np.allclose(setup.g.xyz[::-1, i], s.xyz[:, i])



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

from sisl import Atom, Geometry, Hamiltonian, HydrogenicOrbital

pytestmark = [pytest.mark.physics]


def test_EigenstateElectron_norm2():
    pz = HydrogenicOrbital(2, 1, 0, 3.2)
    C1 = Atom(6, (pz))
    C2 = Atom(6, (pz, pz))
    na = 4
    xyz = np.zeros((na, 3))
    xyz[:, 0] = 1.42 * np.arange(na)
    g = Geometry(xyz, atoms=(C1, C2))
    H = Hamiltonian(g)
    for i in range(H.no - 1):
        H[i, i + 1] = -2.7
        H[i + 1, i] = -2.7

    state = H.eigenstate()
    assert len(state) == H.no
    assert state.norm2()[0] == pytest.approx(1)
    assert state.norm2().shape == (H.no,)
    for p in ("diagonal", "orbital", "atom"):
        assert state.norm2(projection=p).sum() == pytest.approx(H.no)

    ns = 3
    state3 = state.sub(range(ns))
    assert state3.norm2(projection="trace").ndim == 0
    assert state3.norm2(projection="diagonal").shape == (ns,)
    assert state3.norm2(projection="orbital").shape == (ns, H.no)
    assert state3.norm2(projection="atom").shape == (ns, H.na)
    assert state3.norm2(projection="atom").sum() == pytest.approx(ns)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

import sisl

pytestmark = [pytest.mark.io, pytest.mark.siesta]


def test_si_pdos_kgrid_grid(sisl_files):
    si = sisl.get_sile(sisl_files("siesta", "Si_pdos_k", "Si_pdos.VT"))
    si.read_grid()
    assert si.grid_unit == pytest.approx(sisl.unit.siesta.unit_convert("Ry", "eV"))


def test_si_pdos_kgrid_grid_cell(sisl_files):
    si = sisl.get_sile(sisl_files("siesta", "Si_pdos_k", "Si_pdos.VT"))
    si.read_lattice()


def test_si_pdos_kgrid_grid_fractions(sisl_files):
    si = sisl.get_sile(sisl_files("siesta", "Si_pdos_k", "Si_pdos.VT"))
    grid = si.read_grid()
    grid_halve = si.read_grid(index=[0.5])
    assert np.allclose(grid.grid * 0.5, grid_halve.grid)


def test_si_pdos_kgrid_grid_fdf(sisl_files):
    si = sisl.get_sile(sisl_files("siesta", "Si_pdos_k", "Si_pdos.fdf"))
    VT = si.read_grid("VT", order="bin")
    TotPot = si.read_grid("totalpotential", order="bin")
    assert np.allclose(VT.grid, TotPot.grid)


def test_grid_read_write(sisl_tmp):
    path = sisl_tmp("grid.bin")
    lat = sisl.Lattice(1).rotate(45, "z").rotate(45, "x")
    grid = sisl.Grid([4, 5, 6], lattice=lat)
    grid.grid = np.random.rand(*grid.shape)
    gridSile = sisl.io.siesta.gridSileSiesta

    sile = gridSile(path)
    grid.write(sile)
    grid2 = sile.read_grid()
    assert np.allclose(grid.shape, grid2.shape)
    assert np.allclose(grid.cell, grid2.cell)
    assert np.allclose(grid.grid, grid2.grid)

    sile = gridSile(path)
    sile.write_grid(grid, grid * 2)
    for idx in (0, 1):
        grid2 = sile.read_grid(index=idx)
        assert np.allclose(grid.shape, grid2.shape)
        assert np.allclose(grid.cell, grid2.cell)
        assert np.allclose(grid.grid * (1 + idx), grid2.grid)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest
from scipy import interpolate as interp

try:
    from scipy.integrate import trapezoid
except ImportError:
    from scipy.integrate import trapz as trapezoid

from sisl._core.orbital import (
    AtomicOrbital,
    GTOrbital,
    HydrogenicOrbital,
    Orbital,
    SphericalOrbital,
    STOrbital,
    _rspher_harm_fact,
)
from sisl.utils.mathematics import cart2spher, spher2cart

pytestmark = [pytest.mark.orbital]

_max_l = len(_rspher_harm_fact) - 1


def r_f(n):
    r = np.arange(n)
    return r, r


def test_spherical():
    rad2 = np.pi / 45
    r, theta, phi = np.ogrid[0.1:10:0.2, -np.pi : np.pi : rad2, 0 : np.pi : rad2]
    xyz = spher2cart(r, theta, phi)
    s = xyz.shape[:-1]
    r1, theta1, phi1 = cart2spher(xyz)
    r1.shape = s
    theta1.shape = s
    phi1.shape = s
    assert np.allclose(r, r1)
    assert np.allclose(theta, theta1[1:2, :, 1:2])
    assert np.allclose(phi, phi1[0:1, 0:1, :])


class Test_orbital:
    def test_init1(self):
        assert Orbital(1.0) == Orbital(1.0)
        assert Orbital(1.0, tag="none") != Orbital(1.0)
        assert Orbital(1.0, 1.0) != Orbital(1.0)
        assert Orbital(1.0, 1.0) != Orbital(1.0, 1.0, tag="none")

    def test_basic1(self):
        orb = Orbital(1.0)
        str(orb)
        orb = Orbital(1.0, tag="none")
        str(orb)
        orb = Orbital(1.0, 1.0, tag="none")
        str(orb)
        assert orb == orb.copy()
        assert orb != 1.0

    def test_copy(self):
        orb = Orbital(1.0)
        assert orb.R == orb.copy().R
        orb = Orbital(-1.0)
        assert orb.R == orb.copy().R

    def test_psi1(self):
        # Orbital does not have radial part
        with pytest.raises(NotImplementedError):
            Orbital(1.0).psi(np.arange(10))

    def test_scale1(self):
        o = Orbital(1.0)
        assert o.scale(2).R == 2.0
        o = Orbital(-1)
        assert o.scale(2).R == -1.0

    def test_pickle1(self):
        import pickle as p

        o0 = Orbital(1.0)
        o1 = Orbital(1.0, tag="none")
        p0 = p.dumps(o0)
        p1 = p.dumps(o1)
        l0 = p.loads(p0)
        l1 = p.loads(p1)
        assert o0 == l0
        assert o1 == l1
        assert o0 != l1
        assert o1 != l0


class Test_sphericalorbital:
    def test_init1(self):
        n = 6
        rf = np.arange(n)
        rf = (rf, rf)
        assert SphericalOrbital(1, rf) == SphericalOrbital(1, rf)
        f = interp.interp1d(
            rf[0], rf[1], fill_value=(0.0, 0.0), bounds_error=False, kind="cubic"
        )
        rf = [rf[0], rf[0]]
        assert SphericalOrbital(1, rf) == SphericalOrbital(1, f)
        assert SphericalOrbital(1, rf, tag="none") != SphericalOrbital(1, rf)
        SphericalOrbital(5, rf)
        for l in range(10):
            o = SphericalOrbital(l, rf)
            assert l == o.l
            assert o == o.copy()

    def test_basic1(self):
        rf = r_f(6)
        orb = SphericalOrbital(1, rf)
        str(orb)
        orb = SphericalOrbital(1, rf, tag="none")
        str(orb)

    def test_copy(self):
        rf = r_f(6)
        orb = SphericalOrbital(1, rf, R=2.0)
        assert orb.R == orb.copy().R
        assert orb.R == pytest.approx(2.0)
        orb = SphericalOrbital(1, rf)
        assert orb.R == orb.copy().R

    def test_set_radial1(self):
        rf = r_f(6)
        o = SphericalOrbital(1, rf)
        with pytest.raises(ValueError):
            o.set_radial(1.0)

    def test_set_radial_none(self):
        rf = r_f(6)
        o = SphericalOrbital(1, rf)
        o.set_radial()
        r = np.linspace(0, 6, 400)
        r0 = o.radial(r)
        assert np.allclose(r0, 0)

    def test_radial1(self):
        rf = r_f(6)
        orb0 = SphericalOrbital(0, rf)
        orb1 = SphericalOrbital(1, rf)
        r = np.linspace(0, 6, 400)
        # Note r > n - 1 should be zero, regardless of the fill-value
        r0 = orb0.radial(r)
        r1 = orb1.radial(r)
        rr = np.stack((r, np.zeros(len(r)), np.zeros(len(r))), axis=1)
        r2 = orb1.radial((rr**2).sum(-1) ** 0.5)
        assert np.allclose(r0, r1)
        assert np.allclose(r0, r2)
        r[r >= rf[0].max()] = 0.0
        assert np.allclose(r0, r)
        assert np.allclose(r1, r)

    def test_psi1(self):
        rf = r_f(6)
        orb0 = SphericalOrbital(0, rf)
        orb1 = SphericalOrbital(1, rf)
        r = np.linspace(0, 6, 333 * 3).reshape(-1, 3)
        p0 = orb0.psi(r)
        p1 = orb1.psi(r)
        assert not np.allclose(p0, p1)
        orb0 = SphericalOrbital(1, rf)
        assert orb0.equal(orb1, radial=True, psi=True)

        for m in range(orb0.l, orb0.l + 1):
            p0 = orb0.psi(r, -1)
            p1 = orb1.psi(r, -1)
            assert np.allclose(p0, p1)
        p0 = orb0.psi(r, -1)
        p1 = orb1.psi(r, 1)
        assert not np.allclose(p0, p1)

    def test_radial_func1(self):
        r = np.linspace(0, 4, 300)
        f = np.exp(-r)
        o = SphericalOrbital(1, (r, f), R=4.0)
        str(o)

        def i_univariate(r, f):
            return interp.UnivariateSpline(r, f, k=5, s=0, ext=1, check_finite=False)

        def i_interp1d(r, f):
            return interp.interp1d(
                r, f, kind="cubic", fill_value=(f[0], 0.0), bounds_error=False
            )

        def i_spline(r, f):
            from functools import partial

            tck = interp.splrep(r, f, k=5, s=0)
            return partial(interp.splev, tck=tck, der=0, ext=1)

        # Interpolation radius
        R = np.linspace(0, 5, 400)

        assert np.allclose(o.radial(r), f)
        f_default = o.radial(R)

        o.set_radial(r, f, interp=i_univariate)
        assert np.allclose(o.radial(r), f)
        f_univariate = o.radial(R)
        o.set_radial(r, f, interp=i_interp1d)
        assert np.allclose(o.radial(r), f)
        f_interp1d = o.radial(R)

        o.set_radial(r, f, interp=i_spline)
        assert np.allclose(o.radial(r), f)
        f_spline = o.radial(R)

        # Checks that they are equal
        assert np.allclose(f_univariate, f_interp1d)
        assert np.allclose(f_univariate, f_spline)
        assert np.allclose(f_univariate, f_default)

    def test_same1(self):
        rf = r_f(6)
        o0 = SphericalOrbital(0, rf)
        o1 = Orbital(o0.R)
        assert o0.equal(o1)
        assert not o0.equal(Orbital(3.0))

    def test_toatomicorbital1(self):
        rf = r_f(6)
        # Check m and l
        for l in range(_max_l + 1):
            orb = SphericalOrbital(l, rf)
            ao = orb.toAtomicOrbital()
            assert len(ao) == 2 * l + 1
            m = -l
            for a in ao:
                assert a.l == orb.l
                assert a.m == m
                m += 1

        orb = SphericalOrbital(1, rf)
        ao = orb.toAtomicOrbital(1)
        assert ao.l == orb.l
        assert ao.m == 1
        ao = orb.toAtomicOrbital(-1)
        assert ao.l == orb.l
        assert ao.m == -1
        ao = orb.toAtomicOrbital(0)
        assert ao.l == orb.l
        assert ao.m == 0
        ao = orb.toAtomicOrbital([0, -1, 1])
        for a in ao:
            assert a.l == orb.l
        assert ao[0].m == 0
        assert ao[1].m == -1
        assert ao[2].m == 1

    def test_toatomicorbital2(self):
        rf = r_f(6)
        orb = SphericalOrbital(1, rf)
        with pytest.raises(ValueError):
            ao = orb.toAtomicOrbital(2)

    def test_toatomicorbital_q0(self):
        rf = r_f(6)
        orb = SphericalOrbital(0, rf, 2.0)

        # Check m and l
        for l in range(_max_l + 1):
            orb = SphericalOrbital(l, rf, 2.0)
            ao = orb.toAtomicOrbital()
            assert ao[0].q0 == pytest.approx(2.0 / (2 * l + 1))

    def test_pickle1(self):
        rf = r_f(6)
        import pickle as p

        o0 = SphericalOrbital(1, rf)
        o1 = SphericalOrbital(2, rf)
        p0 = p.dumps(o0)
        p1 = p.dumps(o1)
        l0 = p.loads(p0)
        l1 = p.loads(p1)
        assert o0 == l0
        assert o1 == l1
        assert o0 != l1
        assert o1 != l0

    def test_togrid1(self):
        o = SphericalOrbital(1, r_f(6))
        o.toGrid()
        o.toGrid(R=10)

    def test_togrid2(self):
        o = SphericalOrbital(1, r_f(6))
        with pytest.raises(ValueError):
            o.toGrid(R=-1)


class Test_atomicorbital:
    def test_init1(self):
        rf = r_f(6)
        a = []
        a.append(AtomicOrbital(2, 1, 0, 1, True, rf))
        a.append(AtomicOrbital(l=1, m=0, zeta=1, P=True, spherical=rf))
        f = interp.interp1d(
            rf[0], rf[1], fill_value=(0.0, 0.0), bounds_error=False, kind="cubic"
        )
        a.append(AtomicOrbital(l=1, m=0, zeta=1, P=True, spherical=f))
        a.append(AtomicOrbital("pzP", f))
        a.append(AtomicOrbital("pzP", rf))
        a.append(AtomicOrbital("2pzP", rf))
        for i in range(len(a) - 1):
            for j in range(i + 1, len(a)):
                assert a[i] == a[j] and a[i].equal(a[j], psi=True, radial=True)

    def test_init2(self):
        assert AtomicOrbital("pzP") == AtomicOrbital(n=2, l=1, m=0, P=True)

    def test_init3(self):
        rf = r_f(6)
        for l in range(_max_l + 1):
            a = AtomicOrbital(l=l, m=0, spherical=rf)
            a.name()
            a.name(True)
            str(a)
            a = AtomicOrbital(l=l, m=0, P=True, spherical=rf, tag="hello")
            a.name()
            a.name(True)
            str(a)

    def test_init4(self):
        rf = r_f(6)
        o1 = AtomicOrbital(2, 1, 0, 1, True, rf)
        o2 = AtomicOrbital("pzP", rf)
        o3 = AtomicOrbital("pzZP", rf)
        o4 = AtomicOrbital("pzZ1P", rf)
        o5 = AtomicOrbital("2pzZ1P", rf)
        assert o1 == o2
        assert o1 == o3
        assert o1 == o4
        assert o1 == o5

    def test_init5(self):
        with pytest.raises(ValueError):
            AtomicOrbital(5, _max_l + 1, 0)

    def test_copy(self):
        rf = r_f(6)
        orb = AtomicOrbital("pzP", rf, R=2.0)
        assert orb.R == orb.copy().R
        assert orb.R == pytest.approx(2.0)
        orb = AtomicOrbital("pzP", rf)
        assert orb.R == orb.copy().R

    def test_radial1(self):
        rf = r_f(6)
        r = np.linspace(0, 6, 100)
        for l in range(_max_l + 1):
            so = SphericalOrbital(l, rf)
            sor = so.radial(r)
            for m in range(-l, l + 1):
                o = AtomicOrbital(l=l, m=m, spherical=rf)
                assert np.allclose(sor, o.radial(r))
                o.set_radial(rf[0], rf[1])
                assert np.allclose(sor, o.radial(r))

    def test_phi1(self):
        rf = r_f(6)
        r = np.linspace(0, 6, 999).reshape(-1, 3)
        for l in range(_max_l + 1):
            so = SphericalOrbital(l, rf)
            for m in range(-l, l + 1):
                o = AtomicOrbital(l=l, m=m, spherical=rf)
                assert np.allclose(so.psi(r, m), o.psi(r))

    def test_pickle1(self):
        import pickle as p

        rf = r_f(6)
        o0 = AtomicOrbital(2, 1, 0, 1, True, rf, tag="hello", q0=1.0)
        o1 = AtomicOrbital(l=1, m=0, zeta=1, P=False, spherical=rf)
        o2 = AtomicOrbital(l=1, m=0, zeta=1, P=False)
        p0 = p.dumps(o0)
        p1 = p.dumps(o1)
        p2 = p.dumps(o2)
        l0 = p.loads(p0)
        l1 = p.loads(p1)
        l2 = p.loads(p2)
        assert o0 == l0
        assert o1 == l1
        assert o2 == l2
        assert o0 != l1
        assert o1 != l0
        assert o2 != l0


class Test_hydrogenicorbital:
    def test_init(self):
        orb = HydrogenicOrbital(2, 1, 0, 3.2)

    def test_basic1(self):
        orb = HydrogenicOrbital(2, 1, 0, 3.2, R=4.0)
        assert orb.R == orb.copy().R
        assert orb.R == pytest.approx(4.0)
        orb = HydrogenicOrbital(2, 1, 0, 3.2)
        assert orb.R == orb.copy().R

    def test_copy(self):
        orb = HydrogenicOrbital(2, 1, 0, 3.2, tag="test", q0=2.5)
        orb2 = orb.copy()
        assert orb.n == orb2.n
        assert orb.l == orb2.l
        assert orb.m == orb2.m
        assert orb.q0 == orb2.q0
        assert orb.tag == orb2.tag

    def test_normalization(self):
        for n in range(6):
            zeff = n * 0.9
            for l in range(n):
                orb = HydrogenicOrbital(n, l, 0, zeff)
                x = np.linspace(0, orb.R, 1000, endpoint=True)
                Rnl = orb.radial(x)
                I = trapezoid(x**2 * Rnl**2, x=x)
                assert abs(I - 1) < 1e-4

    def test_togrid(self):
        for n in range(3):
            zeff = n * 0.9
            for l in range(n):
                for m in range(-l, l + 1):
                    orb = HydrogenicOrbital(n, l, m, zeff)
                    g = orb.toGrid(0.1)
                    I = (g.grid**2).sum() * g.dvolume
                    assert abs(I - 1) < 1e-3

    def test_pickle(self):
        import pickle as p

        o0 = HydrogenicOrbital(2, 1, 0, 3.2, tag="test", q0=2.5)
        o1 = HydrogenicOrbital(2, 1, 0, 3.2)
        p0 = p.dumps(o0)
        p1 = p.dumps(o1)
        l0 = p.loads(p0)
        l1 = p.loads(p1)
        assert o0 == l0
        assert o1 == l1
        assert o0 != l1
        assert o1 != l0


class Test_GTO:
    def test_init(self):
        alpha = [1, 2]
        coeff = [0.1, 0.44]
        orb = GTOrbital(2, 1, 0, alpha, coeff)
        assert orb.R > 0

    def test_copy(self):
        alpha = [1, 2]
        coeff = [0.1, 0.44]
        orb = GTOrbital(2, 1, 0, alpha, coeff, R=4.0)
        assert orb.R == orb.copy().R
        assert orb.R == pytest.approx(4.0)
        orb = GTOrbital(2, 1, 0, alpha, coeff)
        assert orb.R == orb.copy().R

    def test_gto_funcs(self):
        alpha = [0.1688, 0.6239, 3.425]
        coeff = [0.4, 0.7, 1.3]
        x = np.linspace(0, 10, 1000)
        orb = GTOrbital(2, 1, 0, alpha, coeff, R=x[-1])
        assert orb.R == pytest.approx(x[-1])
        Rnl = orb.radial(x)

        R = np.random.rand(10, 3)
        orb.psi(R)

        theta, phi = np.random.rand(2, 10)
        orb.spher(theta, phi)

        orb.psi_spher((R**2).sum(-1) ** 0.5, theta, phi)


class Test_STO:
    def test_init(self):
        alpha = [1, 2]
        coeff = [0.1, 0.44]
        orb = STOrbital(2, 1, 0, alpha, coeff)
        assert orb.R > 0

    def test_copy(self):
        alpha = [1, 2]
        coeff = [0.1, 0.44]
        orb = STOrbital(2, 1, 0, alpha, coeff, R=4.0)
        assert orb.R == orb.copy().R
        assert orb.R == pytest.approx(4.0)
        orb = STOrbital(2, 1, 0, alpha, coeff)
        assert orb.R == orb.copy().R

    def test_sto_funcs(self):
        alpha = [0.1688, 0.6239, 3.425]
        coeff = [0.4, 0.7, 1.3]
        x = np.linspace(0, 10, 1000)
        orb = STOrbital(2, 1, 0, alpha, coeff, R=x[-1])
        assert orb.R == pytest.approx(x[-1])
        Rnl = orb.radial(x)

        R = np.random.rand(10, 3)
        orb.psi(R)

        theta, phi = np.random.rand(2, 10)
        orb.spher(theta, phi)

        orb.psi_spher((R**2).sum(-1) ** 0.5, theta, phi)
